import numpy as np

from ... import ml
from . import SeparationException, MaskSeparationBase


ALLOWED_CLUSTERING_TYPES = ['KMeans', 'GaussianMixture', 'MiniBatchKMeans']


class ClusteringSeparationBase(MaskSeparationBase):
    """
    A base class for any clustering-based separation approach. Subclasses 
    of this class must implement just one function to use it: `extract_features`.
    This function should uses the internal variables of the class to 
    extract the appropriate time-frequency features of the signal. These 
    time-frequency features will then be clustered by `cluster_features`. 
    Masks will then be produced by the run function and applied to the 
    audio signal to produce separated estimates.
    
    Args:
        input_audio_signal: (`AudioSignal`) An AudioSignal object containing the 
          mixture to be separated.
          
        num_sources (int): Number of sources to cluster the features of and separate
          the mixture.

        clustering_type (str): One of 'KMeans', 'GaussianMixture', and 'MiniBatchKMeans'.
          The clustering approach to use on the features. Defaults to 'KMeans'.

        fit_clusterer (bool, optional): Whether or not to call fit on the clusterer.
          If False, then the clusterer should already be fit for this to work. Defaults
          to True.

        percentile (int, optional): Percentile of time-frequency points to consider by loudness. 
          Audio spectrograms are very high dimensional, and louder points tend to 
          matter more than quieter points. By setting the percentile high, one can more
          efficiently cluster an auditory scene by considering only points above
          that threshold. Defaults to 90 (which means the top 10 percentile of 
          time-frequency points will be used for clustering).

        beta (float, optional): When using KMeans, we use soft KMeans, which has an additional 
          parameter `beta`. `beta` controls how soft the assignments are. As beta 
          increases, the assignments become more binary (either 0 or 1). Defaults to 
          5.0, a value discovered through cross-validation.

        mask_type (str, optional): Masking approach to use. Passed up to MaskSeparationBase.

        mask_threshold (float, optional): Threshold for masking. Passed up to MaskSeparationBase.

        **kwargs (dict, optional): Additional keyword arguments that are passed to the clustering
          object (one of KMeans, GaussianMixture, or MiniBatchKMeans).
    
    Raises:
        SeparationException: If clustering type is not one of the allowed ones, or if
          the output of `extract_features` has the wrong shape according to the STFT
          shape of the AudioSignal.
    """
    def __init__(self, input_audio_signal, num_sources, clustering_type='KMeans', 
                 fit_clusterer=True, percentile=90, beta=5.0, mask_type='soft',
                 mask_threshold=0.5, **kwargs):

        if clustering_type not in dir(ml.cluster):
            raise SeparationException(
                f"clustering_type = {clustering_type} not allowed!" 
                f"Use one of {ALLOWED_CLUSTERING_TYPES}."
            )

        ClusterClass = getattr(ml.cluster, clustering_type)

        self.num_sources = num_sources
        if clustering_type == 'GaussianMixture':
            self.clusterer = ClusterClass(n_components=num_sources, **kwargs)
        else:
            self.clusterer = ClusterClass(n_clusters=num_sources, **kwargs)
        self.clustering_type = clustering_type

        init = None
        if 'init' in kwargs and not fit_clusterer:
            init = kwargs['init']
            self.clusterer.cluster_centers_ = init

        self.percentile = percentile
        self.features = None
        self.beta = beta
        self.fit_clusterer = fit_clusterer

        super(ClusteringSeparationBase, self).__init__(
            input_audio_signal=input_audio_signal,
            mask_type=mask_type,
            mask_threshold=mask_threshold
        )

        self.metadata.update({
            'clustering_type': clustering_type,
            'percentile': percentile,
            'beta': beta,
            'fit_clusterer': fit_clusterer,
            'init': init
        })

    def _preprocess_audio_signal(self):
        """
        Preprocess the audio signal object - takes STFTs, sets features to None 
        (new audio signal means new features), finds the time frequency points above
        the cutoff according to the percentile.
        """
        self.features = None
        self.stft = self.audio_signal.stft()

        # get a cutoff using the percentile
        self.cutoff = np.percentile(np.abs(self.stft), self.percentile)
        self.tf_point_over_cutoff = np.abs(self.stft) >= self.cutoff

    def confidence(self, approach='silhouette_confidence', **kwargs):
        """
        In clustering-based separation algorithms, we can compute a confidence
        measure based on the clusterability of the feature space. This can be computed
        only after the features have been extracted by ``extract_features``. 
        
        Args:
            approach (str, optional): What approach to use for getting the confidence 
              measure. Options are 'jensen_shannon_confidence', 'posterior_confidence', 
              'silhouette_confidence', 'loudness_confidence', 'whitened_kmeans_confidence',
              'dpcl_classic_confidence'. Defaults to 'silhouette_confidence'.
            kwargs: Keyword arguments to the function being used to compute the confidence.
        """
        if self.features is None:
            raise SeparationException(
                "self.features is None! Did you run extract features?")
        confidence_function = getattr(ml.confidence, approach)
        confidence = confidence_function(
            self.audio_signal, self.features, self.num_sources, **kwargs)
        return confidence


    def extract_features(self):
        """
        This function should be implemented by the subclass. It should extract
        features. If the STFT shape is `(n_freq, n_time, n_chan)`, the output of this
        function should be `(n_freq, n_time, n_chan, n_features)`.
        """
        raise NotImplementedError()

    def cluster_features(self, features, clusterer):
        """
        Clusters each time-frequency point according to features for each
        time-frequency point. Features should be on the last axis.

        Features should come in in the shape:
          `(..., n_features)`
        
        Args:
            features (np.ndarray): Features to cluster, for each time-frequency point.
            clusterer (object): Clustering object to use.
        
        Returns:
            np.ndarray: Responsibilities for each cluster for each time-frequency point. 
        """
        shape = features.shape
        features_to_fit = features.reshape(-1, shape[-1])
        features_to_fit = features_to_fit[self.tf_point_over_cutoff.flatten(), :]
        if self.fit_clusterer:
            clusterer.fit(features_to_fit)
        responsibilities = None

        if 'KMeans' in self.clustering_type:
            distances = clusterer.transform(features.reshape(-1, shape[-1]))
            distances = np.exp(-self.beta * distances) + 1e-8
            responsibilities = distances / distances.sum(axis=-1, keepdims=True)
        elif 'GaussianMixture' in self.clustering_type:
            responsibilities = clusterer.predict_proba(features.reshape(-1, shape[-1]))

        responsibilities = responsibilities.reshape(shape[:-1] + (-1,))
        return responsibilities

    def run(self, features=None):
        """
        Clusters the features using the chosen clustering algorithm.
        
        Args:
            features (np.ndarray, optional): If features are given, then the 
              `extract_features` step will be skipped. Defaults to None (so
              `extract_features` will be run.)
        
        Raises:
            SeparationException: If features.shape doesn't match what is expected
              in the STFT of the audio signal, an exception is raised.
        
        Returns:
            list: List of Mask objects in self.result_masks.
        """
        if features is None:
            features = self.extract_features()

        if features.shape[:-1] != self.stft.shape:
            raise SeparationException(
                f"features.shape did not match stft shape along all but feature " 
                f"dimension! Got {features.shape[:-1]}, expected {self.stft.shape}.")

        responsibilities = self.cluster_features(features, self.clusterer)
        self.features = features
        self.result_masks = []

        for i in range(responsibilities.shape[-1]):
            mask_data = responsibilities[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = (
                    responsibilities[..., i] == responsibilities.max(axis=-1))
            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)
        
        return self.result_masks



import copy
import warnings

import numpy as np

from ... import AudioSignal, play_utils


class SeparationBase(object):
    """Base class for all separation algorithms in nussl.

    Do not call this. It will not do anything.

    Parameters:
        input_audio_signal (AudioSignal). AudioSignal` object.
            This will always be a copy of the provided AudioSignal object.
    """

    def __init__(self, input_audio_signal):
        self.metadata = {}
        self._audio_signal = None
        self.audio_signal = input_audio_signal

    @property
    def sample_rate(self):
        """
        (int): Sample rate of :attr:`audio_signal`.
        Literally :attr:`audio_signal.sample_rate`.
        """
        return self.audio_signal.sample_rate

    @property
    def stft_params(self):
        """
        STFTParams object containing the STFT parameters of the copied AudioSignal.
        """
        return self.audio_signal.stft_params

    @property
    def audio_signal(self):
        """
        Copy of AudioSignal that is made on initialization.
        """
        return self._audio_signal

    def _preprocess_audio_signal(self):
        """
        This function should be implemented by the subclass. It can do things like
        take the STFT of the audio signal, or resample it to a desired sample rate,
        build the input data for a deep model, etc. Here, it does nothing.
        """
        pass

    @audio_signal.setter
    def audio_signal(self, input_audio_signal):
        """
        When setting the AudioSignal object for a separation algorithm (which
        can happen on initialization or later one), it is copied on set so
        as to not alter the data within the original audio signal. If the
        AudioSignal object has data, then it the function `_preprocess_audio_signal`
        is run, which is implemented by the subclass. 
        
        Args:
            input_audio_signal ([type]): [description]
        """
        if not isinstance(input_audio_signal, AudioSignal):
            raise ValueError('input_audio_signal is not an AudioSignal object!')

        self._audio_signal = copy.deepcopy(input_audio_signal)

        if self.audio_signal is not None:
            if not self.audio_signal.has_data:
                warnings.warn('input_audio_signal has no data!')

                # initialize to empty arrays so that we don't crash randomly
                self.audio_signal.audio_data = np.array([])
                self.audio_signal.stft_data = np.array([[]])
            else:
                self._preprocess_audio_signal()

    def interact(self, add_residual=False, source='upload', label=None, share=False):
        """
        Uses gradio to create a small interactive interface
        for the separation algorithm. Fair warning, there
        may be some race conditions with this...

        When you call this from a notebook, the interface will be displayed
        below the cell. When you call this from a regular Python script, you'll see a 
        link print out (a localhost link and a gradio link if you
        called this with sharing on). The sessions will last for the duration
        of the notebook or the script.

        To use this functionality, you must install gradio: `pip install gradio`.

        Args:
            add_residual: Whether or not to add the residual signal.
            source: Either "upload" (upload a file to separate), or "microphone", record.
            share: Whether or not to create a public gradio link.
            kwargs: Keyword arguments to gradio.

        Example:
        
            >>> import nussl
            >>> nussl.separation.primitive.HPSS(
            >>>     nussl.AudioSignal()).interact()

        """
        try:
            import gradio
        except: # pragma: no cover
            raise ImportError(
                "To use this functionality, you must install gradio: "
                "pip install gradio.")
                
        def _separate(file_obj): # pragma: no cover
            mix = AudioSignal(file_obj.name)
            self.audio_signal = mix
            estimates = self()
            if add_residual:
                estimates.append(mix - estimates[0])

            estimates = {f'Estimate {i}': s for i, s in enumerate(estimates)}
            html = play_utils.multitrack(estimates, ext='.mp3', display=False)
            
            return html

        if label is None: label = f"Separation via {type(self).__name__}"

        audio_in = gradio.inputs.Audio(source=source, type="file", label=label)

        gradio.Interface(
            fn=_separate, 
            inputs=audio_in, 
            outputs="html",
        ).launch(share=share)

    def run(self, *args, audio_signal=None, **kwargs):
        """
        Runs separation algorithm.

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def make_audio_signals(self):
        """
        Makes :class:`audio_signal.AudioSignal` objects after separation algorithm is run

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def get_metadata(self, to_str=False, **kwargs):
        """
        Returns metadata associated with this separation algorithm.

        Args:
            to_str (bool): Whether to return the metadata as a string.

        Returns:
            Formatted metadata if `to_str` is True, else metadata dict.

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def __call__(self, *args, audio_signal=None, **kwargs):
        if audio_signal is not None:
            self.audio_signal = audio_signal
        
        self.run(*args, **kwargs)
        return self.make_audio_signals()

    def __repr__(self):
        return f"{self.__class__.__name__} on {str(self.audio_signal)}"

    def __eq__(self, other):
        for k, v in list(self.__dict__.items()):
            if isinstance(v, np.ndarray):
                if not np.array_equal(v, other.__dict__[k]):
                    return False
            elif v != other.__dict__[k]:
                return False
        return True

    def __ne__(self, other):
        return not self == other


class SeparationException(Exception):
    pass



import os
import shutil
import logging
import random
from collections import OrderedDict

import torch
import zarr
import numcodecs
import numpy as np
from sklearn.preprocessing import OneHotEncoder

from .. import utils

# This is for when you're running multiple
# training threads
if hasattr(numcodecs, 'blosc'):
    numcodecs.blosc.use_threads = False


def compute_ideal_binary_mask(source_magnitudes):
    ibm = (
            source_magnitudes == np.max(source_magnitudes, axis=-1, keepdims=True)
    ).astype(float)

    ibm = ibm / np.sum(ibm, axis=-1, keepdims=True)
    ibm[ibm <= .5] = 0
    return ibm


# Keys that correspond to the time-frequency representations after being passed through
# the transforms here.
time_frequency_keys = ['mix_magnitude', 'source_magnitudes', 'ideal_binary_mask', 'weights']


class SumSources(object):
    """
    Sums sources together. Looks for sources in ``data[self.source_key]``. If 
    a source belongs to a group, it is popped from the ``data[self.source_key]`` and
    summed with the other sources in the group. If there is a corresponding 
    group_name in group_names, it is named that in ``data[self.source_key]``. If
    group_names are not given, then the names are constructed using the keys
    in each group (e.g. `drums+bass+other`).

    If using Scaper datasets, then there may be multiple sources with the same
    label but different counts. The Scaper dataset hook organizes the source
    dictionary as follows:

    .. code-block:: none

        data['sources] = {
            '{label}::{count}': AudioSignal,
            '{label}::{count}': AudioSignal,
            ...
        }
    
    SumSources sums by source label, so the ``::count`` will be ignored and only the
    label part will be used when grouping sources.

    Example:
        >>> import nussl
        >>> tfm = nussl.datasets.transforms.SumSources(
                groupings=[['drums', 'bass', 'other]],
                group_names=['accompaniment],
            )
        >>> # data['sources'] is a dict containing keys: 
        >>> #   ['vocals', 'drums', 'bass', 'other]
        >>> data = tfm(data)
        >>> # data['sources'] is now a dict containing keys:
        >>> #   ['vocals', 'accompaniment']
    
    Args:
        groupings (list): a list of lists telling how to group each sources. 
        group_names (list, optional): A list containing the names of each group, or None. 
          Defaults to None.
        source_key (str, optional): The key to look for in the data containing the list of
          source AudioSignals. Defaults to 'sources'.
    
    Raises:
        TransformException: if groupings is not a list
        TransformException: if group_names is not None but 
            len(groupings) != len(group_names)
    
    Returns:
        data: modified dictionary with summed sources
    """

    def __init__(self, groupings, group_names=None, source_key='sources'):
        if not isinstance(groupings, list):
            raise TransformException(
                f"groupings must be a list, got {type(groupings)}!")

        if group_names:
            if len(group_names) != len(groupings):
                raise TransformException(
                    f"group_names and groupings must be same length or "
                    f"group_names can be None! Got {len(group_names)} for "
                    f"len(group_names) and {len(groupings)} for len(groupings)."
                )

        self.groupings = groupings
        self.source_key = source_key
        if group_names is None:
            group_names = ['+'.join(groupings[i]) for i in range(len(groupings))]
        self.group_names = group_names

    def __call__(self, data):
        if self.source_key not in data:
            raise TransformException(
                f"Expected {self.source_key} in dictionary "
                f"passed to this Transform!"
            )
        sources = data[self.source_key]
        source_keys = [(k.split('::')[0], k) for k in list(sources.keys())]

        for i, group in enumerate(self.groupings):
            combined = []
            group_name = self.group_names[i]
            for key1 in group:
                for key2 in source_keys:
                    if key2[0] == key1:
                        combined.append(sources[key2[1]])
                        sources.pop(key2[1])
            sources[group_name] = sum(combined)
            sources[group_name].path_to_input_file = group_name

        data[self.source_key] = sources
        if 'metadata' in data:
            if 'labels' in data['metadata']:
                data['metadata']['labels'].extend(self.group_names)

        return data

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"groupings = {self.groupings}, "
            f"group_names = {self.group_names}, "
            f"source_key = {self.source_key}"
            f")"
        )


class LabelsToOneHot(object):
    """
    Takes a data dictionary with sources and their keys and converts the keys to
    a one-hot numpy array using the list in data['metadata']['labels'] to figure
    out which index goes where.
    """

    def __init__(self, source_key='sources'):
        self.source_key = source_key

    def __repr__(self):
        return f'{self.__class__.__name__}(source_key = {self.source_key})'

    def __call__(self, data):
        if 'metadata' not in data:
            raise TransformException(
                f"Expected metadata in data, got {list(data.keys())}")
        if 'labels' not in data['metadata']:
            raise TransformException(
                f"Expected labels in data['metadata'], got "
                f"{list(data['metadata'].keys())}")

        enc = OneHotEncoder(categories=[data['metadata']['labels']])

        sources = data[self.source_key]
        source_keys = [k.split('::')[0] for k in list(sources.keys())]
        source_labels = [[l] for l in sorted(source_keys)]

        one_hot_labels = enc.fit_transform(source_labels)
        data['one_hot_labels'] = one_hot_labels.toarray()

        return data


class MagnitudeSpectrumApproximation(object):
    """
    Takes a dictionary and looks for two special keys, defined by the
    arguments ``mix_key`` and ``source_key``. These default to `mix` and `sources`.
    These values of these keys are used to calculate the magnitude spectrum 
    approximation [1]. The input dictionary is modified to have additional
    keys:

    - mix_magnitude: The magnitude spectrogram of the mixture audio signal.
    - source_magnitudes: The magnitude spectrograms of each source spectrogram.
    - assignments: The ideal binary assignments for each time-frequency bin.

    ``data[self.source_key]`` points to a dictionary containing the source names in
    the keys and the corresponding AudioSignal in the values. The keys are sorted
    in alphabetical order and then appended to the mask. ``data[self.source_key]``
    then points to an OrderedDict instead, where the keys are in the same order
    as in ``data['source_magnitudes']`` and ``data['assignments']``.

    This transform uses the STFTParams that are attached to the AudioSignal objects
    contained in ``data[mix_key]`` and ``data[source_key]``.

    [1] Erdogan, Hakan, John R. Hershey, Shinji Watanabe, and Jonathan Le Roux. 
        "Phase-sensitive and recognition-boosted speech separation using 
        deep recurrent neural networks." In 2015 IEEE International Conference 
        on Acoustics, Speech and Signal Processing (ICASSP), pp. 708-712. IEEE, 
        2015.
    
    Args:
        mix_key (str, optional): The key to look for in data for the mixture AudioSignal. 
          Defaults to 'mix'.
        source_key (str, optional): The key to look for in the data containing the dict of
          source AudioSignals. Defaults to 'sources'.
    
    Raises:
            TransformException: if the expected keys are not in the dictionary, an
              Exception is raised.
        
    Returns:
        data: Modified version of the input dictionary.
    """

    def __init__(self, mix_key='mix', source_key='sources'):
        self.mix_key = mix_key
        self.source_key = source_key

    def __call__(self, data):
        if self.mix_key not in data:
            raise TransformException(
                f"Expected {self.mix_key} in dictionary "
                f"passed to this Transform! Got {list(data.keys())}."
            )

        mixture = data[self.mix_key]
        mixture.stft()
        mix_magnitude = mixture.magnitude_spectrogram_data

        data['mix_magnitude'] = mix_magnitude

        if self.source_key not in data:
            return data

        _sources = data[self.source_key]
        source_names = sorted(list(_sources.keys()))

        sources = OrderedDict()
        for key in source_names:
            sources[key] = _sources[key]
        data[self.source_key] = sources

        source_magnitudes = []
        for key in source_names:
            s = sources[key]
            s.stft()
            source_magnitudes.append(s.magnitude_spectrogram_data)

        source_magnitudes = np.stack(source_magnitudes, axis=-1)

        data['ideal_binary_mask'] = compute_ideal_binary_mask(source_magnitudes)
        data['source_magnitudes'] = source_magnitudes

        return data

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"mix_key = {self.mix_key}, "
            f"source_key = {self.source_key}"
            f")"
        )


class MagnitudeWeights(object):
    """
    Applying time-frequency weights to the deep clustering objective results in a
    huge performance boost. This transform looks for 'mix_magnitude', which is output
    by either MagnitudeSpectrumApproximation or PhaseSensitiveSpectrumApproximation
    and puts it into the weights.

    [1] Wang, Zhong-Qiu, Jonathan Le Roux, and John R. Hershey. 
    "Alternative objective functions for deep clustering." 2018 IEEE International 
    Conference on Acoustics, Speech and Signal Processing (ICASSP). IEEE, 2018.
    
    Args:
        mix_magnitude_key (str): Which key to look for the mix_magnitude data in.
    """

    def __init__(self, mix_key='mix', mix_magnitude_key='mix_magnitude'):
        self.mix_magnitude_key = mix_magnitude_key
        self.mix_key = mix_key

    def __repr__(self):
        return f'{self.__class__.__name__}(mix_key = {self.mix_key}, ' \
               f'mix_magnitude_key = {self.mix_magnitude_key})'

    def __call__(self, data):
        if self.mix_magnitude_key not in data and self.mix_key not in data:
            raise TransformException(
                f"Expected {self.mix_magnitude_key} or {self.mix_key} in dictionary "
                f"passed to this Transform! Got {list(data.keys())}. "
                "Either MagnitudeSpectrumApproximation or "
                "PhaseSensitiveSpectrumApproximation should be called "
                "on the data dict prior to this transform. "
            )
        elif self.mix_magnitude_key not in data:
            data[self.mix_magnitude_key] = np.abs(data[self.mix_key].stft())

        magnitude_spectrogram = data[self.mix_magnitude_key]
        weights = magnitude_spectrogram / (np.sum(magnitude_spectrogram) + 1e-6)
        weights *= (
                magnitude_spectrogram.shape[0] * magnitude_spectrogram.shape[1]
        )
        data['weights'] = np.sqrt(weights)
        return data


class PhaseSensitiveSpectrumApproximation(object):
    """
    Takes a dictionary and looks for two special keys, defined by the
    arguments ``mix_key`` and ``source_key``. These default to `mix` and `sources`.
    These values of these keys are used to calculate the phase sensitive spectrum 
    approximation [1]. The input dictionary is modified to have additional
    keys:

    - mix_magnitude: The magnitude spectrogram of the mixture audio signal.
    - source_magnitudes: The magnitude spectrograms of each source spectrogram.
    - assignments: The ideal binary assignments for each time-frequency bin.

    ``data[self.source_key]`` points to a dictionary containing the source names in
    the keys and the corresponding AudioSignal in the values. The keys are sorted
    in alphabetical order and then appended to the mask. ``data[self.source_key]``
    then points to an OrderedDict instead, where the keys are in the same order
    as in ``data['source_magnitudes']`` and ``data['assignments']``.

    This transform uses the STFTParams that are attached to the AudioSignal objects
    contained in ``data[mix_key]`` and ``data[source_key]``.

    [1] Erdogan, Hakan, John R. Hershey, Shinji Watanabe, and Jonathan Le Roux. 
        "Phase-sensitive and recognition-boosted speech separation using 
        deep recurrent neural networks." In 2015 IEEE International Conference 
        on Acoustics, Speech and Signal Processing (ICASSP), pp. 708-712. IEEE, 
        2015.
    
    Args:
        mix_key (str, optional): The key to look for in data for the mixture AudioSignal. 
          Defaults to 'mix'.
        source_key (str, optional): The key to look for in the data containing the list of
          source AudioSignals. Defaults to 'sources'.
        range_min (float, optional): The lower end to use when truncating the source 
          magnitudes in the phase sensitive spectrum approximation. Defaults to 0.0 (construct
          non-negative masks). Use -np.inf for untruncated source magnitudes.
        range_max (float, optional): The higher end of the truncated spectrum. This gets
          multiplied by the magnitude of the mixture. Use 1.0 to truncate the source 
          magnitudes to `max(source_magnitudes, mix_magnitude)`. Use np.inf for untruncated
          source magnitudes (best performance for an oracle mask but may be beyond what a
          neural network is capable of masking). Defaults to 1.0.
          
    Raises:
            TransformException: if the expected keys are not in the dictionary, an
              Exception is raised.
        
    Returns:
        data: Modified version of the input dictionary.
    """

    def __init__(self, mix_key='mix', source_key='sources',
                 range_min=0.0, range_max=1.0):
        self.mix_key = mix_key
        self.source_key = source_key
        self.range_min = range_min
        self.range_max = range_max

    def __call__(self, data):
        if self.mix_key not in data:
            raise TransformException(
                f"Expected {self.mix_key} in dictionary "
                f"passed to this Transform! Got {list(data.keys())}."
            )

        mixture = data[self.mix_key]

        mix_stft = mixture.stft()
        mix_magnitude = np.abs(mix_stft)
        mix_angle = np.angle(mix_stft)
        data['mix_magnitude'] = mix_magnitude

        if self.source_key not in data:
            return data

        _sources = data[self.source_key]
        source_names = sorted(list(_sources.keys()))

        sources = OrderedDict()
        for key in source_names:
            sources[key] = _sources[key]
        data[self.source_key] = sources

        source_angles = []
        source_magnitudes = []
        for key in source_names:
            s = sources[key]
            _stft = s.stft()
            source_magnitudes.append(np.abs(_stft))
            source_angles.append(np.angle(_stft))

        source_magnitudes = np.stack(source_magnitudes, axis=-1)
        source_angles = np.stack(source_angles, axis=-1)
        range_min = self.range_min
        range_max = self.range_max * mix_magnitude[..., None]

        # Section 3.1: https://arxiv.org/pdf/1909.08494.pdf
        source_magnitudes = np.minimum(
            np.maximum(
                source_magnitudes * np.cos(source_angles - mix_angle[..., None]),
                range_min
            ),
            range_max
        )

        data['ideal_binary_mask'] = compute_ideal_binary_mask(source_magnitudes)
        data['source_magnitudes'] = source_magnitudes

        return data

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"mix_key = {self.mix_key}, "
            f"source_key = {self.source_key}, "
            f"range_min = {self.range_min}, "
            f"range_max = {self.range_max}"
            f")"
        )


class IndexSources(object):
    """
    Takes in a dictionary containing Torch tensors or numpy arrays and extracts the 
    indexed sources from the set key (usually either `source_magnitudes` or 
    `ideal_binary_mask`). Can be used to train single-source separation models 
    (e.g. mix goes in, vocals come out).
    
    You need to know which slice of the source magnitudes or ideal binary mask arrays 
    to extract. The order of the sources in the source magnitudes array will be in 
    alphabetical order according to their source labels. 

    For example, if source magnitudes has shape `(257, 400, 1, 4)`, and the data is
    from MUSDB, then the four possible source labels are bass, drums, other, and vocals.
    The data in source magnitudes is in alphabetical order, so:

    .. code-block:: python

        # source_magnitudes is an array returned by either MagnitudeSpectrumApproximation
        # or PhaseSensitiveSpectrumApproximation
        source_magnitudes[..., 0] # bass spectrogram
        source_magnitudes[..., 1] # drums spectrogram
        source_magnitudes[..., 2] # other spectrogram
        source_magnitudes[..., 3] # vocals spectrogram

        # ideal_binary_mask is an array returned by either MagnitudeSpectrumApproximation
        # or PhaseSensitiveSpectrumApproximation
        ideal_binary_mask[..., 0] # bass ibm mask
        ideal_binary_mask[..., 1] # drums ibm mask
        ideal_binary_mask[..., 2] # other ibm mask
        ideal_binary_mask[..., 3] # vocals ibm mask

    You can apply this transform to either the `source_magnitudes` or the
    `ideal_binary_mask` or both.

    
    Args:
        object ([type]): [description]
    """

    def __init__(self, target_key, index):
        self.target_key = target_key
        self.index = index

    def __call__(self, data):
        if self.target_key not in data:
            raise TransformException(
                f"Expected {self.target_key} in dictionary, got {list(data.keys())}")
        if self.index >= data[self.target_key].shape[-1]:
            raise TransformException(
                f"Shape of data[{self.target_key}] is {data[self.target_key].shape} "
                f"but index =  {self.index} out of bounds bounds of last dim.")
        data[self.target_key] = data[self.target_key][..., self.index, None]
        return data

    def __repr__(self):
        return f'{self.__class__.__name__}(target_key = {self.target_key}, ' \
               f'index = {self.index})'


class GetExcerpt(object):
    """
    Takes in a dictionary containing Torch tensors or numpy arrays and extracts an 
    excerpt from each tensor corresponding to a spectral representation of a specified 
    length in frames. Can be used to get L-length spectrograms from mixture and source
    spectrograms. If the data is shorter than the specified length, it
    is padded to the specified length. If it is longer, a random offset between
    ``(0, data_length - specified_length)`` is chosen. This function assumes that
    it is being passed data AFTER ToSeparationModel. Thus the time dimension is
    on axis=1. 
    
    Args:
        excerpt_length (int): Specified length of transformed data in frames.

        time_dim (int): Which dimension time is on (excerpts are taken along this axis).
          Defaults to 0.

        time_frequency_keys (list): Which keys to look at it in the data dictionary to
          take excerpts from.
    """

    def __init__(self, excerpt_length, time_dim=0,
                 tf_keys=None):
        self.excerpt_length = excerpt_length
        self.time_dim = time_dim
        self.time_frequency_keys = tf_keys if tf_keys else time_frequency_keys

    def __repr__(self):
        return f'{self.__class__.__name__}(excerpt_length = {self.excerpt_length}), ' \
               f'time_dim = {self.time_dim}, tf_keys = {self.time_frequency_keys})'

    @staticmethod
    def _validate(data, key):
        is_tensor = torch.is_tensor(data[key])
        is_array = isinstance(data[key], np.ndarray)
        if not is_tensor and not is_array:
            raise TransformException(
                f"data[{key}] was not a torch Tensor or a numpy array!")
        return is_tensor, is_array

    def _get_offset(self, data, key):
        self._validate(data, key)
        data_length = data[key].shape[self.time_dim]

        if data_length >= self.excerpt_length:
            offset = random.randint(0, data_length - self.excerpt_length)
        else:
            offset = 0

        pad_amount = max(0, self.excerpt_length - data_length)

        return offset, pad_amount

    def _construct_pad_func_tuple(self, shape, pad_amount, is_tensor):
        if is_tensor:
            pad_func = torch.nn.functional.pad
            pad_tuple = [0 for _ in range(2 * len(shape))]
            pad_tuple[2 * self.time_dim] = pad_amount
            pad_tuple = pad_tuple[::-1]
        else:
            pad_func = np.pad
            pad_tuple = [(0, 0) for _ in range(len(shape))]
            pad_tuple[self.time_dim] = (0, pad_amount)
        return pad_func, pad_tuple

    def __call__(self, data):
        offset, pad_amount = self._get_offset(
            data, self.time_frequency_keys[0])

        for key in data:
            if key in self.time_frequency_keys:
                is_tensor, is_array = self._validate(data, key)

                if pad_amount > 0:
                    pad_func, pad_tuple = self._construct_pad_func_tuple(
                        data[key].shape, pad_amount, is_tensor)
                    data[key] = pad_func(data[key], pad_tuple)

                data[key] = utils._slice_along_dim(
                    data[key], self.time_dim, offset, offset + self.excerpt_length)
        return data


class Cache(object):
    """
    The Cache transform can be placed within a Compose transform. The data 
    dictionary coming into this transform will be saved to the specified
    location using ``zarr``. Then instead of computing all of the transforms
    before the cache, one can simply read from the cache. The transforms after
    this will then be applied to the data dictionary that is read from the
    cache. A typical pipeline might look like this:

    .. code-block:: python

        dataset = datasets.Scaper('path/to/scaper/folder')
        tfm = transforms.Compose([
            transforms.PhaseSensitiveApproximation(),
            transforms.ToSeparationModel(),
            transforms.Cache('~/.nussl/cache/tag', overwrite=True),
            transforms.GetExcerpt()
        ])
        dataset[0] # first time will write to cache then apply GetExcerpt
        dataset.cache_populated = True # switches to reading from cache
        dataset[0] # second time will read from cache then apply GetExcerpt
        dataset[1] # will error out as it wasn't written to the cache!

        dataset.cache_populated = False
        for i in range(len(dataset)):
            dataset[i] # every item will get written to cache
        dataset.cache_populated = True
        dataset[1] # now it exists

        dataset = datasets.Scaper('path/to/scaper/folder') # next time around
        tfm = transforms.Compose([
            transforms.PhaseSensitiveApproximation(),
            transforms.ToSeparationModel(),
            transforms.Cache('~/.nussl/cache/tag', overwrite=False),
            transforms.GetExcerpt()
        ])
        dataset.cache_populated = True
        dataset[0] # will read from cache, which still exists from last time
    
    Args:
        object ([type]): [description]
    """

    def __init__(self, location, cache_size=1, overwrite=False):
        self.location = location
        self.cache_size = cache_size
        self.cache = None
        self.overwrite = overwrite

    @property
    def info(self):
        return self.cache.info

    @property
    def overwrite(self):
        return self._overwrite

    @overwrite.setter
    def overwrite(self, value):
        self._overwrite = value
        self._clear_cache(self.location)
        self._open_cache(self.location)

    def _clear_cache(self, location):
        if os.path.exists(location):
            if self.overwrite:
                logging.info(
                    f"Cache {location} exists and overwrite = True, clearing cache.")
                shutil.rmtree(location, ignore_errors=True)

    def _open_cache(self, location):
        if self.overwrite:
            self.cache = zarr.open(location, mode='w', shape=(self.cache_size,),
                                   chunks=(1,), dtype=object, 
                                   object_codec=numcodecs.Pickle(),
                                   synchronizer=zarr.ThreadSynchronizer())
        else:
            if os.path.exists(location):
                self.cache = zarr.open(location, mode='r',
                    object_codec=numcodecs.Pickle(),
                    synchronizer=zarr.ThreadSynchronizer())

    def __call__(self, data):
        if 'index' not in data:
            raise TransformException(
                f"Expected 'index' in dictionary, got {list(data.keys())}")
        index = data['index']
        if self.overwrite:
            self.cache[index] = data
        data = self.cache[index]

        if not isinstance(data, dict):
            raise TransformException(
                f"Reading from cache resulted in not a dictionary! "
                f"Maybe you haven't written to index {index} yet in "
                f"the cache?")

        return data


class GetAudio(object):
    """
    Extracts the audio from each signal in `mix_key` and `source_key`. 
    These will be at new keys, called `mix_audio` and `source_audio`.
    Can be used for training end-to-end models.
    
    Args:
        mix_key (str, optional): The key to look for in data for the mixture AudioSignal. 
          Defaults to 'mix'.
        source_key (str, optional): The key to look for in the data containing the dict of
          source AudioSignals. Defaults to 'sources'.
    """
    def __init__(self, mix_key='mix', source_key='sources'):
        self.mix_key = mix_key
        self.source_key = source_key

    def __repr__(self):
        return f'{self.__class__.__name__}(mix_key = {self.mix_key}, ' \
               f'source_key = {self.source_key})'

    def __call__(self, data):
        if self.mix_key not in data:
            raise TransformException(
                f"Expected {self.mix_key} in dictionary "
                f"passed to this Transform! Got {list(data.keys())}."
            )
        
        mix = data[self.mix_key]
        data['mix_audio'] = mix.audio_data

        if self.source_key not in data:
            return data

        _sources = data[self.source_key]
        source_names = sorted(list(_sources.keys()))
        
        source_audio = []
        for key in source_names:
            source_audio.append(_sources[key].audio_data)
        # sources on last axis
        source_audio = np.stack(source_audio, axis=-1)

        data['source_audio'] = source_audio
        return data


class ToSeparationModel(object):
    """
    Takes in a dictionary containing objects and removes any objects that cannot
    be passed to SeparationModel (e.g. not a numpy array or torch Tensor).
    If these objects are passed to SeparationModel, then an error will occur. This 
    class should be the last one in your list of transforms, if you're using 
    this dataset in a DataLoader object for training a network. If the keys
    correspond to numpy arrays, they are converted to tensors using 
    ``torch.from_numpy``. Finally, the dimensions corresponding to time and
    frequency are swapped for all the keys in swap_tf_dims, as this is how 
    SeparationModel expects it.

    Example:

    .. code-block:: none

        data = {
            # 2ch spectrogram for mixture
            'mix_magnitude': torch.randn(513, 400, 2),
            # 2ch spectrogram for each source
            'source_magnitudes': torch.randn(513, 400, 2, 4) 
            'mix': AudioSignal()
        }

        tfm = transforms.ToSeparationModel()
        data = tfm(data)

        data['mix_magnitude'].shape # (400, 513, 2)
        data['source_magnitudes].shape # (400, 513, 2, 4)
        'mix' in data.keys() # False
    

    If this class isn't in your transforms list for the dataset, but you are
    using it in the Trainer class, then it is added automatically as the
    last transform.
    """

    def __init__(self, swap_tf_dims=None):
        self.swap_tf_dims = swap_tf_dims if swap_tf_dims else time_frequency_keys

    def __call__(self, data):
        keys = list(data.keys())
        for key in keys:
            if key != 'index':
                is_array = isinstance(data[key], np.ndarray)
                if is_array:
                    data[key] = torch.from_numpy(data[key])
                if not torch.is_tensor(data[key]):
                    data.pop(key)
                if key in self.swap_tf_dims:
                    data[key] = data[key].transpose(1, 0)
        return data

    def __repr__(self):
        return f"{self.__class__.__name__}(swap_tf_dims = {self.swap_tf_dims})"


class Compose(object):
    """Composes several transforms together. Inspired by torchvision implementation.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Example:
        >>> transforms.Compose([
        >>>     transforms.MagnitudeSpectrumApproximation(),
        >>>     transforms.ToSeparationModel(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, data):
        for t in self.transforms:
            data = t(data)
            if not isinstance(data, dict):
                raise TransformException(
                    "The output of every transform must be a dictionary!")
        return data

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string


class TransformException(Exception):
    """
    Exception class for errors when working with transforms in nussl.
    """
    pass






import numpy as np

from .. import MaskSeparationBase
from ..benchmark import HighLowPassFilter


class RPCA(MaskSeparationBase):
    """
    Implements foreground/background separation using RPCA.

    Huang, Po-Sen, et al. "Singing-voice separation from monaural recordings using 
    robust principal component analysis." Acoustics, Speech and Signal Processing 
    (ICASSP), 2012 IEEE International Conference on. IEEE, 2012.

    Args:
        input_audio_signal (AudioSignal): The AudioSignal object that has the
          audio data that RPCA will be run on.
        high_pass_cutoff (float, optional): Value (in Hz) for the high pass cutoff 
          filter. Defaults to 100.
        num_iterations (int, optional): how many iterations to run RPCA for. 
          Defaults to 100.
        epsilon (float, optional): Stopping criterion for RPCA convergence. 
          Defaults to 1e-7.
        mask_type (str, optional): Type of mask to use. Defaults to 'soft'.
        mask_threshold (float, optional): Threshold for mask. Defaults to 0.5.
    """

    def __init__(self, input_audio_signal, high_pass_cutoff=100, num_iterations=100, 
                 epsilon=1e-7, mask_type='soft', mask_threshold=0.5):
        super().__init__(
            input_audio_signal=input_audio_signal, 
            mask_type=mask_type,
            mask_threshold=mask_threshold)
        self.high_pass_cutoff = high_pass_cutoff

        self.epsilon = epsilon
        self.num_iterations = num_iterations
        self.gain = 1

        self.error = None
        self.magnitude_spectrogram = None

    def run(self):
        high_low = HighLowPassFilter(self.audio_signal, self.high_pass_cutoff)
        high_pass_masks = high_low.run()

        self.magnitude_spectrogram = np.abs(self.stft)

        background_masks = []
        foreground_masks = []

        for ch in range(self.audio_signal.num_channels):
            background_mask = self._compute_rpca_mask(
                self.magnitude_spectrogram[..., ch])
            foreground_mask = 1 - background_mask
            
            background_masks.append(background_mask)
            foreground_masks.append(foreground_mask)

        background_masks = np.stack(background_masks, axis=-1)
        foreground_masks = np.stack(foreground_masks, axis=-1)

        _masks = np.stack([background_masks, foreground_masks], axis=-1)

        self.result_masks = []

        for i in range(_masks.shape[-1]):
            mask_data = _masks[..., i]
            if self.mask_type == self.MASKS['binary']:
                mask_data = _masks[..., i] == np.max(_masks, axis=-1)
            
            if i == 0:
                mask_data = np.maximum(mask_data, high_pass_masks[i].mask)
            elif i == 1:
                mask_data = np.minimum(mask_data, high_pass_masks[i].mask)
            
            mask = self.mask_type(mask_data)
            self.result_masks.append(mask)

        return self.result_masks

    def _compute_rpca_mask(self, magnitude_spectrogram):
        low_rank, sparse_matrix = self.decompose(magnitude_spectrogram)
        bg_mask = self.gain * np.abs(sparse_matrix) <= np.abs(low_rank)
        return bg_mask.astype(float)

    def decompose(self, magnitude_spectrogram):
        # compute rule of thumb values of lagrange multiplier and svd-threshold
        _lambda = 1 / np.sqrt(np.max(magnitude_spectrogram.shape))

        # initialize low rank and sparse matrices to all zeros
        low_rank = np.zeros(magnitude_spectrogram.shape)
        sparse_matrix = np.zeros(magnitude_spectrogram.shape)

        # get singular values for magnitude_spectrogram
        two_norm = np.linalg.svd(magnitude_spectrogram, full_matrices=False, compute_uv=False)[0]
        inf_norm = np.linalg.norm(magnitude_spectrogram.flatten(), np.inf) / _lambda
        dual_norm = np.max([two_norm, inf_norm])
        residuals = magnitude_spectrogram / dual_norm

        # tunable parameters
        mu = 1.25 / two_norm
        mu_bar = mu * 1e7
        rho = 1.5

        error = np.inf
        converged = False
        num_iteration = 0

        while not converged and num_iteration < self.num_iterations:
            num_iteration += 1
            low_rank = self.svd_threshold(magnitude_spectrogram - sparse_matrix + residuals / mu,
                                          1 / mu)
            sparse_matrix = self.shrink(magnitude_spectrogram - low_rank + residuals / mu,
                                        _lambda / mu)
            residuals += mu * (magnitude_spectrogram - low_rank - sparse_matrix)
            mu = np.min([mu * rho, mu_bar])
            error = np.linalg.norm(magnitude_spectrogram - low_rank - sparse_matrix,
                                   ord='fro') / np.linalg.norm(magnitude_spectrogram, ord='fro')
            if error < self.epsilon:
                converged = True
        self.error = error
        return low_rank, sparse_matrix

    @staticmethod
    def shrink(matrix, tau):
        return np.sign(matrix) * np.maximum(np.abs(matrix) - tau, 0)

    def svd_threshold(self, matrix, tau):
        u, sigma, v = np.linalg.svd(matrix, full_matrices=False)
        shrunk = self.shrink(sigma, tau)
        thresholded_singular_values = np.dot(u, np.dot(np.diag(shrunk), v))
        return thresholded_singular_values



"""
High pass filter
----------------

.. autoclass:: nussl.separation.benchmark.HighLowPassFilter
    :autosummary:

Ideal binary mask
-----------------

.. autoclass:: nussl.separation.benchmark.IdealBinaryMask
    :autosummary:

Ideal ratio mask
----------------

.. autoclass:: nussl.separation.benchmark.IdealRatioMask
    :autosummary:

Wiener filter
-------------

.. autoclass:: nussl.separation.benchmark.WienerFilter
    :autosummary:

Mix as estimate
---------------

.. autoclass:: nussl.separation.benchmark.MixAsEstimate
    :autosummary:

"""

from .high_low_pass_filter import HighLowPassFilter
from .ideal_binary_mask import IdealBinaryMask
from .ideal_ratio_mask import IdealRatioMask
from .wiener_filter import WienerFilter
from .mix_as_estimate import MixAsEstimate



import warnings
from typing import Iterable
import copy

from torch.utils.data import Dataset

from .. import AudioSignal
from . import transforms as tfm
import tqdm


class BaseDataset(Dataset, Iterable):
    """
    The BaseDataset class is the starting point for all dataset hooks
    in nussl. To subclass BaseDataset, you only have to implement two 
    functions:

    - ``get_items``: a function that is passed the folder and generates a
      list of items that will be processed by the next function. The
      number of items in the list will dictate len(dataset). Must return
      a list.
    - ``process_item``: this function processes a single item in the list
      generated by get_items. Must return a dictionary.

    After process_item is called, a set of Transforms can be applied to the 
    output of process_item. If no transforms are defined (``self.transforms = None``),
    then the output of process_item is returned by self[i]. For implemented
    Transforms, see nussl.datasets.transforms. For example, 
    PhaseSpectrumApproximation will add three new keys to the output dictionary
    of process_item:

    - mix_magnitude: the magnitude spectrogram of the mixture
    - source_magnitudes: the magnitude spectrogram of each source
    - ideal_binary_mask: the ideal binary mask for each source

    The transforms are applied in sequence using transforms.Compose. 
    Not all sequences of transforms will be valid (e.g. if you pop a key in
    one transform but a later transform operates on that key, you will get
    an error).

    For examples of subclassing, see ``nussl.datasets.hooks``.
    
    Args:
        folder (str): location that should be processed to produce the list of files

        transform (transforms.* object, optional): A transforms to apply to the output of
          ``self.process_item``. If using transforms.Compose, each transform will be
          applied in sequence. Defaults to None.

        sample_rate (int, optional): Sample rate to use for each audio files. If
          audio file sample rate doesn't match, it will be resampled on the fly.
          If None, uses the default sample rate. Defaults to None.

        stft_params (STFTParams, optional): STFTParams object defining window_length,
          hop_length, and window_type that will be set for each AudioSignal object. 
          Defaults to None (32ms window length, 8ms hop, 'hann' window).

        num_channels (int, optional): Number of channels to make each AudioSignal
          object conform to. If an audio signal in your dataset has fewer channels
          than ``num_channels``, a warning is raised, as the behavior in this case
          is undefined. Defaults to None.

        strict_sample_rate (bool, optional): Whether to raise an error if 
    
    Raises:
        DataSetException: Exceptions are raised if the output of the implemented
            functions by the subclass don't match the specification.
    """
    def __init__(self, folder, transform=None, sample_rate=None, stft_params=None,
                 num_channels=None, strict_sample_rate=True, cache_populated=False):
        self.folder = folder
        self.items = self.get_items(self.folder)
        self.transform = transform

        self.cache_populated = cache_populated
        
        self.stft_params = stft_params
        self.sample_rate = sample_rate
        self.num_channels = num_channels
        self.strict_sample_rate = strict_sample_rate
        self.metadata = {
            'name': self.__class__.__name__,
            'stft_params': stft_params,
            'sample_rate': sample_rate,
            'num_channels': num_channels,
            'folder': folder,
            'transforms': copy.deepcopy(transform)
        }

        if not isinstance(self.items, list):
            raise DataSetException("Output of self.get_items must be a list!")
        
        # getting one item in order to set up parameters for audio
        # signals if necessary, if there are any items
        if self.items:
            self.process_item(self.items[0])

    def filter_items_by_condition(self, func):
        """
        Filter the items in the list according to a function that takes 
        in both the dataset as well as the item currently be processed.
        If the item in the list passes the condition, then it is kept
        in the list. Otherwise it is taken out of the list. For example,
        a function that would get rid of an item if it is below some 
        minimum number of seconds would look like this:

        .. code-block:: python

            min_length = 1 # in seconds

            # self here refers to the dataset
            def remove_short_audio(self, item):
                processed_item = self.process_item(item)
                mix_length = processed_item['mix'].signal_duration
                if mix_length < min_length:
                    return False
                return True
            
            dataset.items # contains all items
            dataset.filter_items_by_condition(remove_short_audio)
            dataset.items # contains only items longer than min length
            
        Args:
            func (function): A function that takes in two arguments: the dataset and
              this dataset object (self). The function must return a bool.
        """
        filtered_items = []
        n_removed = 0
        desc = f"Filtered {n_removed} items out of dataset"
        pbar = tqdm.tqdm(self.items, desc=desc)
        for item in pbar:
            check = func(self, item)
            if not isinstance(check, bool):
                raise DataSetException(
                    "Output of filter function must be True or False!"
                )
            if check:
                filtered_items.append(item)
            else:
                n_removed += 1
            pbar.set_description(f"Filtered {n_removed} items out of dataset")
        self.items = filtered_items

    @property
    def cache_populated(self):
        return self._cache_populated

    @cache_populated.setter
    def cache_populated(self, value):
        self.post_cache_transforms = []
        cache_transform = None

        transforms = (
            self.transform.transforms 
            if isinstance(self.transform, tfm.Compose) 
            else [self.transform])

        found_cache_transform = False
        for t in transforms:
            if isinstance(t, tfm.Cache):
                found_cache_transform = True
                cache_transform = t
            if found_cache_transform:
                self.post_cache_transforms.append(t)

        if not found_cache_transform:
            # there is no cache transform
            self._cache_populated = False
        else:
            self._cache_populated = value
            cache_transform.cache_size = len(self)
            cache_transform.overwrite = not value

            self.post_cache_transforms = tfm.Compose(
                self.post_cache_transforms)

    def get_items(self, folder):
        """
        This function must be implemented by whatever class inherits BaseDataset.
        It should return a list of items in the given folder, each of which is 
        processed by process_items in some way to produce mixes, sources, class
        labels, etc.

        Args:
            folder (str): location that should be processed to produce the list of files.

        Returns:
            list: list of items that should be processed
        """
        raise NotImplementedError()

    def __len__(self):
        """
        Gets the length of the dataset (the number of items that will be processed).

        Returns:
            int: Length of the dataset (``len(self.items)``).
        """
        return len(self.items)

    def __getitem__(self, i):
        """
        Processes a single item in ``self.items`` using ``self.process_item``.
        The output of ``self.process_item`` is further passed through bunch of
        of transforms if they are defined in parallel. If you want to have
        a set of transforms that depend on each other, then you should compose them
        into a single transforms and then pass it into here. The output of each
        transform is added to an output dictionary which is returned by this
        function.
        
        Args:
            i (int): Index of the dataset to return. Indexes ``self.items``.

        Returns:
            dict: Dictionary with keys and values corresponding to the processed
                item after being put through the set of transforms (if any are
                defined).
        """
        if self.cache_populated:
            data = {'index': i}
            data = self.post_cache_transforms(data)
        else:
            data = self.process_item(self.items[i])

            if not isinstance(data, dict):
                raise DataSetException(
                    "The output of process_item must be a dictionary!")

            if self.transform:
                data['index'] = i
                data = self.transform(data)

                if not isinstance(data, dict):
                    raise tfm.TransformException(
                        "The output of transform must be a dictionary!")

        return data

    def __iter__(self):
        """
        Calls ``self.__getitem__`` from ``0`` to ``self.__len__()``.
        Required when inheriting Iterable.

        Yields:
            dict: Dictionary with keys and values corresponding to the processed
                item after being put through the set of transforms (if any are
                defined).
        """
        for i in range(len(self)):
            yield self[i]

    def process_item(self, item):
        """Each file returned by get_items is processed by this function. For example,
        if each file is a json file containing the paths to the mixture and sources, 
        then this function should parse the json file and load the mixture and sources
        and return them.

        Exact behavior of this functionality is determined by implementation by subclass.

        Args:
            item (object): the item that will be processed by this function. Input depends
              on implementation of ``self.get_items``.

        Returns:
            This should return a dictionary that gets processed by the transforms.
        """
        raise NotImplementedError()

    def _load_audio_file(self, path_to_audio_file, **kwargs):
        """
        Loads audio file at given path. Uses AudioSignal to load the audio data
        from disk.

        Args:
            path_to_audio_file: relative or absolute path to file to load
            kwargs: Keyword arguments to AudioSignal.

        Returns:
            AudioSignal: loaded AudioSignal object of path_to_audio_file
        """
        audio_signal = AudioSignal(path_to_audio_file, **kwargs)
        self._setup_audio_signal(audio_signal)
        return audio_signal
    
    def _load_audio_from_array(self, audio_data, sample_rate=None):
        """
        Loads the audio data into an AudioSignal object with the appropriate 
        sample rate.
        
        Args:
            audio_data (np.ndarray): numpy array containing the samples containing
              the audio data.

            sample_rate (int): the sample rate at which to load the audio file. 
              If None, self.sample_rate or the sample rate of the actual file is used. 
              Defaults to None.
        
        Returns:
            AudioSignal: loaded AudioSignal object of audio_data
        """
        sample_rate = sample_rate if sample_rate else self.sample_rate
        audio_signal = AudioSignal(
            audio_data_array=audio_data, sample_rate=sample_rate)
        self._setup_audio_signal(audio_signal)
        return audio_signal

    def _setup_audio_signal(self, audio_signal):
        """
        You will want every item from a dataset to be uniform in sample rate, STFT
        parameters, and number of channels. This function takes an audio signal 
        object loaded by the dataset and uses it to set the sample rate, STFT parameters,
        and the number of channels. If ``self.sample_rate``, ``self.stft_params``, and
        ``self.num_channels`` are set at construction time of the dataset, then the
        opposite happens - attributes of the AudioSignal object are set to the desired
        values.
        
        Args:
            audio_signal (AudioSignal): AudioSignal object to query to set the parameters
              of this dataset or to set the parameters of, according to what is in the 
              dataset.
        """
        if self.sample_rate and self.sample_rate != audio_signal.sample_rate:
            if self.strict_sample_rate:
                raise DataSetException(
                    f"All audio files should have been the same sample rate already " 
                    f"because self.strict_sample_rate = True. Please resample or "
                    f"turn set self.strict_sample_rate = False"
                )
            audio_signal.resample(self.sample_rate)  
        else:
            self.sample_rate = audio_signal.sample_rate

        # set audio signal attributes to requested values, if they exist
        if self.stft_params:
            audio_signal.stft_params = self.stft_params
        else:
            self.stft_params = audio_signal.stft_params

        if self.num_channels:
            if audio_signal.num_channels > self.num_channels:
                # pick the first ``self.num_channels`` channels 
                audio_signal.audio_data = audio_signal.audio_data[:self.num_channels]
            elif audio_signal.num_channels < self.num_channels:
                warnings.warn(
                    f"AudioSignal had {audio_signal.num_channels} channels "
                    f"but self.num_channels = {self.num_channels}. Unsure "
                    f"of what to do, so warning. You might want to make sure "
                    f"your dataset is uniform!"
                )
        else:
            self.num_channels = audio_signal.num_channels

class DataSetException(Exception):
    """
    Exception class for errors when working with data sets in nussl.
    """
    pass



"""
Separation algorithms
=====================

Base classes
------------

These classes are used to build every type of source separation
algorithm currently in nussl. They provide helpful utilities
and make it such that the end-user only has to implement
one or two functions to create a new separation algorithm,
depending on what sort of algorithm they are trying to 
implement.

.. automodule:: nussl.separation.base
    :members:
    :autosummary:
    :undoc-members:

Benchmark methods
-----------------

These methods are used for obtaining upper and lower baselines
for source separation algorithms.

.. automodule:: nussl.separation.benchmark
    :members:
    :autosummary:
    :undoc-members:

Deep methods
------------

Deep networks can be used for source separation via these
classes.

.. automodule:: nussl.separation.deep
    :members:
    :autosummary:
    :undoc-members:

Composite methods
-----------------

These are methods that use the output of 
multiple separation algorithms to build better
more robust separation estimates.

.. automodule:: nussl.separation.composite
    :members:
    :autosummary:
    :undoc-members:

Factorization-based methods
---------------------------

The methods use some sort of factorization-based algorithm
like robust principle component analysis, or independent
component analysis to separate the auditory scene.

.. automodule:: nussl.separation.factorization
    :members:
    :autosummary:
    :undoc-members:

Primitive methods
-----------------

These methods are based on primitives - hard-wired perceptual
grouping cues that are used automatically by the brain. 
Primitives were coined by Albert Bregman in the book 
*Auditory Scene Analysis*. 

.. automodule:: nussl.separation.primitive
    :members:
    :autosummary:
    :undoc-members:

Spatial methods
---------------

These methods are based on primitive spatial cues. 

.. automodule:: nussl.separation.spatial
    :members:
    :autosummary:
    :undoc-members:

"""

from .base import (
    SeparationBase, 
    MaskSeparationBase,
    ClusteringSeparationBase,
    SeparationException,
    NMFMixin,
    DeepMixin
)

from . import (
    deep, 
    spatial, 
    benchmark,
    primitive,
    factorization,
    composite
)



import copy
import numbers
import os.path
import warnings
from collections import namedtuple

import audioread
import librosa
import numpy as np
import scipy.io.wavfile as wav
import scipy
from scipy.signal import check_COLA
import soundfile as sf
import pyloudnorm

from . import constants
from . import utils
from . import masks
from . import effects

__all__ = ['AudioSignal', 'STFTParams', 'AudioSignalException']

STFTParams = namedtuple('STFTParams',
                        ['window_length', 'hop_length', 'window_type']
                        )
STFTParams.__new__.__defaults__ = (None,) * len(STFTParams._fields)
"""
STFTParams object is a container that holds STFT parameters - window_length, 
hop_length, and window_type. Not all parameters need to be specified. Ones that
are not specified will be inferred by the AudioSignal parameters and the settings
in `nussl.core.constants`.
"""


class AudioSignal(object):
    """

    **Overview**

    :class:`AudioSignal` is the main entry and exit point for all source separation algorithms
    in *nussl*. The :class:`AudioSignal` class is a general container for all things related to
    audio data. It contains utilities for:

    * Input and output from an array or from a file,
    * Time-series and frequency domain manipulation,
    * Plotting and visualizing,
    * Playing audio within a terminal or jupyter notebook,
    * Applying a mask to estimate signals

    and more. The :class:`AudioSignal` class is used in all source separation objects in *nussl*.

    :class:`AudioSignal` object stores time-series audio data as a 2D ``numpy`` array in
    :attr:`audio_data` (see :attr:`audio_data` for details) and stores Short-Time Fourier Transform
    data as 3D ``numpy`` array in :ref:`stft_data` (see :attr:`stft_data` for details).


    **Initialization**

    There are a few options for initializing an :class:`AudioSignal` object. The first is to
    initialize an empty :class:`AudioSignal` object, with no parameters:

     >>> import nussl
     >>> signal = nussl.AudioSignal()

    In this case, there is no data stored in :attr:`audio_data` or in :attr:`stft_data`, though
    these attributes can be updated at any time after the object has been created.

    Additionally, an :class:`AudioSignal` object can be loaded with exactly one of the following:

        1. A path to an input audio file (see :func:`load_audio_from_file` for details).
        2. A `numpy` array of 1D or 2D real-valued time-series audio data.
        3. A `numpy` array of 2D or 3D complex-valued time-frequency STFT data.

    :class:`AudioSignal` will throw an error if it is initialized with more than one of the
    previous at once.

    Here are examples of all three of these cases:

     .. code-block:: python
        :linenos:

        import numpy as np
        import nussl

        # Initializing an empty AudioSignal object:
        sig_empty = nussl.AudioSignal()

        # Initializing from a path:
        file_path = 'my/awesome/mixture.wav'
        sig_path = nussl.AudioSignal(file_path)

        # Initializing with a 1D or 2D numpy array containing audio data:
        aud_1d = np.sin(np.linspace(0.0, 1.0, 48000))
        sig_1d = nussl.AudioSignal(audio_data_array=aud_1d, sample_rate=48000)

        # FYI: The shape doesn't matter, nussl will correct for it
        aud_2d = np.array([aud_1d, -2 * aud_1d])
        sig_2d = nussl.AudioSignal(audio_data_array=aud_2d)

        # Initializing with a 2D or 3D numpy array containing STFT data:
        stft_2d = np.random.rand((513, 300)) + 1j * np.random.rand((513, 300))
        sig_stft_2d = nussl.AudioSignal(stft=stft_2d)

        # Two channels of STFT data:
        stft_3d = nussl.utils.complex_randn((513, 300, 2))
        sig_stft_3d = nussl.AudioSignal(stft=stft_3d)

        # Initializing with more than one of the above methods will raise an exception:
        sig_exception = nussl.AudioSignal(audio_data_array=aud_2d, stft=stft_2d)

    When initializing from a path, :class:`AudioSignal` can read many types of audio files,
    provided that your computer has the backends installed to understand the corresponding codecs.
    *nussl* uses ``librosa``'s `load` function to read in audio data. See librosa's documentation
    for details: https://github.com/librosa/librosa#audioread

    Once initialized with a single type of data (time-series or time-frequency), there are methods
    to compute an STFT from time-series data (:func:`stft`) and vice versa (:func:`istft`).

    **Sample Rate**

    The sample rate of an :class:`AudioSignal` object is set upon initialization. If initializing
    from a path, the sample rate of the :class:`AudioSignal` object inherits the native sample
    rate from the file. If initialized with an audio or stft data array, the sample rate is passed
    in as an optional argument. In these cases, with no sample rate explicitly defined, the default
    sample rate is 44.1 kHz (CD quality). If this argument is provided when reading from a file
    and the provided sample rate does not match the native sample rate of the file,
    :class:`AudioSignal` will resample the data from the file so that it matches the provided
    sample rate.

    Notes:
        There is no guarantee that data in :attr:`audio_data` corresponds to data in
        :attr:`stft_data`. E.g., when an :class:`AudioSignal` object is initialized with
        :attr:`audio_data` of an audio mixture, its :attr:`stft_data` is ``None`` until :func:`stft`
        is called. Once :func:`stft` is called and a mask is applied to :attr:`stft_data` (via some
        algorithm), the :attr:`audio_data` in this :class:`AudioSignal` object still contains data
        from the original mixture that it was initialized with even though :attr:`stft_data`
        contains altered data. (To hear the results, simply call :func:`istft` on the
        :class:`AudioSignal` object.) It is up to the user to keep track of the contents of
        :attr:`audio_data` and :attr:`stft_data`.

    See Also:
        For a walk-through of AudioSignal features, see :ref:`audio_signal_basics` and
        :ref:`audio_signal_stft`.

    Arguments:
        path_to_input_file (``str``): Path to an input file to load upon initialization. Audio
            gets loaded into :attr:`audio_data`.
        audio_data_array (:obj:`np.ndarray`): 1D or 2D numpy array containing a real-valued,
            time-series representation of the audio.
        stft (:obj:`np.ndarray`): 2D or 3D numpy array containing pre-computed complex-valued STFT
            data.
        label (``str``): A label for this :class:`AudioSignal` object.
        offset (``float``): Starting point of the section to be extracted (in seconds) if
            initializing from  a file.
        duration (``float``): Length of the signal to read from the file (in seconds). Defaults to
            full length of the signal (i.e., ``None``).
        sample_rate (``int``): Sampling rate of this :class:`AudioSignal` object.

    Attributes:
        path_to_input_file (``str``): Path to the input file. ``None`` if this AudioSignal never
            loaded a file, i.e., initialized with a ``np.ndarray``.
        label (``str``): A user-definable label for this :class:`AudioSignal` object.
        applied_effects (``list`` of ``effects.FilterFunction``): Effects applied to this 
        :class:`AudioSignal` object. For more information, see apply_effects. 
        effects_chain (``list`` of ``effects.FilterFunction``): Effects queues to be applied to 
        this :class:`AudioSignal` object. For more information, see apply_effects. 
  
    """

    def __init__(self, path_to_input_file=None, audio_data_array=None, stft=None, label=None,
                 sample_rate=None, stft_params=None, offset=0, duration=None):

        self.path_to_input_file = path_to_input_file
        self._audio_data = None
        self.original_signal_length = None
        self._stft_data = None
        self._sample_rate = None
        self._active_start = None
        self._active_end = None
        self.label = label

        # Assert that this object was only initialized in one way
        got_path = path_to_input_file is not None
        got_audio_array = audio_data_array is not None
        got_stft = stft is not None
        init_inputs = np.array([got_path, got_audio_array, got_stft])

        # noinspection PyPep8
        if len(init_inputs[init_inputs == True]) > 1:  # ignore inspection for clarity
            raise AudioSignalException('Can only initialize AudioSignal object with one and only '
                                       'one of {path, audio, stft}!')

        if path_to_input_file is not None:
            self.load_audio_from_file(self.path_to_input_file, offset, duration, sample_rate)
        elif audio_data_array is not None:
            self.load_audio_from_array(audio_data_array, sample_rate)

        if self._sample_rate is None:
            self._sample_rate = constants.DEFAULT_SAMPLE_RATE \
                if sample_rate is None else sample_rate

        self.stft_data = stft  # complex spectrogram data
        self.stft_params = stft_params

        # Effects
        self._effects_chain = []
        self._effects_applied = []

    def __str__(self):
        dur = f'{self.signal_duration:0.3f}' if self.signal_duration else '[unknown]'
        return (
            f"{self.__class__.__name__} "
            f"({self.label if self.label else 'unlabeled'}): "
            f"{dur} sec @ "
            f"{self.path_to_input_file if self.path_to_input_file else 'path unknown'}, "
            f"{self.sample_rate if self.sample_rate else '[unknown]'} Hz, "
            f"{self.num_channels if self.num_channels else '[unknown]'} ch."
        )

    ##################################################
    #                 Properties
    ##################################################

    @property
    def signal_length(self):
        """
        ``int``
            Number of samples in the active region of :attr:`audio_data`.
            The length of the audio signal represented by this object in samples.

        See Also:
            * :func:`signal_duration` for the signal duration in seconds.
            * :func:`set_active_region_to_default` for information about active regions.
        """
        if self.audio_data is None:
            return self.original_signal_length
        return self.audio_data.shape[constants.LEN_INDEX]

    @property
    def signal_duration(self):
        """
        ``float``
            Duration of the active region of :attr:`audio_data` in seconds.
            The length of the audio signal represented by this object in seconds.

        See Also:
            * :func:`signal_length` for the signal length in samples.
            * :func:`set_active_region_to_default` for information about active regions.
        """
        if self.signal_length is None:
            return None
        return self.signal_length / self.sample_rate

    @property
    def num_channels(self):
        """
        ``int``
            Number of channels this :class:`AudioSignal` has.
            Defaults to returning number of channels in :attr:`audio_data`. If that is ``None``,
            returns number of channels in :attr:`stft_data`. If both are ``None`` then returns
            ``None``.

        See Also:
            * :func:`is_mono`
            * :func:`is_stereo`
        """
        # TODO: what about a mismatch between audio_data and stft_data??
        if self.audio_data is not None:
            return self.audio_data.shape[constants.CHAN_INDEX]
        if self.stft_data is not None:
            return self.stft_data.shape[constants.STFT_CHAN_INDEX]
        return None

    @property
    def is_mono(self):
        """
        ``bool``
            Whether or not this signal is mono (i.e., has exactly **one** channel). First
            looks at :attr:`audio_data`, then (if that's ``None``) looks at :attr:`stft_data`.

        See Also:
            * :func:`num_channels`
            * :func:`is_stereo`
        """
        return self.num_channels == 1

    @property
    def is_stereo(self):
        """
        ``bool``
            Whether or not this signal is stereo (i.e., has exactly **two** channels). First
            looks at :attr:`audio_data`, then (if that's ``None``) looks at :attr:`stft_data`.

        See Also:
            * :func:`num_channels`
            * :func:`is_mono`
        """
        return self.num_channels == 2

    @property
    def audio_data(self):
        """
        ``np.ndarray``
            Stored as a ``numpy`` :obj:`np.ndarray`, :attr:`audio_data` houses the raw, uncompressed
            time-domain audio data in the :class:`AudioSignal`. Audio data is stored with shape
            ``(n_channels, n_samples)`` as an array of floats.

            ``None`` by default, can be initialized upon object instantiation or set at any time by
            accessing this attribute or calling :func:`load_audio_from_array`. It is recommended to
            set :attr:`audio_data` by using :func:`load_audio_from_array` if this
            :class:`AudioSignal` has been initialized without any audio or STFT data.

        Raises:
            :class:`AudioSignalException`
                If set incorrectly, will raise an error. Expects a real, finite-valued 1D or 2D
                ``numpy`` :obj:`np.ndarray`-typed array.

        Warnings:
            :attr:`audio_data` and :attr:`stft_data` are not automatically synchronized, meaning
            that if one of them is changed, those changes are not instantly reflected in the other.
            To propagate changes, either call :func:`stft` or :func:`istft`.


        Notes:
            * This attribute only returns values within the active region. For more information
                see :func:`set_active_region_to_default`. When setting this attribute, the active
                region are reset to default.

            * If :attr:`audio_data` is set with an improperly transposed array, it will
                automatically transpose it so that it is set the expected way. A warning will be
                displayed on the console.

        See Also:
            * :func:`load_audio_from_file` to load audio into :attr:`audio_data` after
                initialization.

            * :func:`load_audio_from_array` to safely load audio into :attr:`audio_data` after
                initialization.

            * :func:`set_active_region_to_default` for more information about the active region.

            * :attr:`signal_duration` and :attr:`signal_length` for length of audio data in seconds
                and samples, respectively.

            * :func:`stft` to calculate an STFT from this data,
                and :func:`istft` to calculate the inverse STFT and put it in :attr:`audio_data`.

            * :func:`plot_time_domain` to create a plot of audio data stored in this attribute.

            * :func:`peak_normalize` to apply gain such that to the absolute max value is exactly
                ``1.0``.

            * :func:`rms` to calculate the root-mean-square of :attr:`audio_data`

            * :func:`apply_gain` to apply a gain.

            * :func:`get_channel` to safely retrieve a single channel in :attr:`audio_data`.
        """
        if self._audio_data is None:
            return None

        start = 0
        end = self._audio_data.shape[constants.LEN_INDEX]

        if self._active_end is not None and self._active_end < end:
            end = self._active_end

        if self._active_start is not None and self._active_start > 0:
            start = self._active_start

        return self._audio_data[:, start:end]

    @audio_data.setter
    def audio_data(self, value):

        if value is None:
            self._audio_data = None
            return

        elif not isinstance(value, np.ndarray):
            raise AudioSignalException('Type of self.audio_data must be of type np.ndarray!')

        if not np.isfinite(value).all():
            raise AudioSignalException('Not all values of audio_data are finite!')

        if value.ndim > 1 and value.shape[constants.CHAN_INDEX] > value.shape[constants.LEN_INDEX]:
            value = value.T

        if value.ndim > 2:
            raise AudioSignalException('self.audio_data cannot have more than 2 dimensions!')

        if value.ndim < 2:
            value = np.expand_dims(value, axis=constants.CHAN_INDEX)

        self._audio_data = value

        self.set_active_region_to_default()

    @property
    def stft_data(self):
        """
        ``np.ndarray``
            Stored as a ``numpy`` :obj:`np.ndarray`, :attr:`stft_data` houses complex-valued data
            computed from a Short-time Fourier Transform (STFT) of audio data in the
            :class:`AudioSignal`. ``None`` by default, this :class:`AudioSignal` object can be
            initialized with STFT data upon initialization or it can be set at any time.

            The STFT data is stored with shape ``(n_frequency_bins, n_hops, n_channels)`` as
            a complex-valued ``numpy`` array.

        Raises:
            :class:`AudioSignalException`
                if set with an :obj:`np.ndarray` with one dimension or more than three dimensions.

        See Also:
            * :func:`stft` to calculate an STFT from :attr:`audio_data`, and :func:`istft` to
             calculate the inverse STFT from this attribute and put it in :attr:`audio_data`.

            * :func:`magnitude_spectrogram` to calculate and get the magnitude spectrogram from
             :attr:`stft_data`. :func:`power_spectrogram` to calculate and get the power
             spectrogram from :attr:`stft_data`.

            * :func:`get_stft_channel` to safely get a specific channel in :attr:`stft_data`.

        Notes:
            * :attr:`audio_data` and :attr:`stft_data` are not automatically synchronized, meaning
            that if one of them is changed, those changes are not instantly reflected in the other.
            To propagate changes, either call :func:`stft` or :func:`istft`.

            * :attr:`stft_data` will expand a two dimensional array so that it has the expected
            shape `(n_frequency_bins, n_hops, n_channels)`.
        """

        return self._stft_data

    @stft_data.setter
    def stft_data(self, value):
        if value is None:
            self._stft_data = None
            return

        elif not isinstance(value, np.ndarray):
            raise AudioSignalException('Type of self.stft_data must be of type np.ndarray!')

        if value.ndim == 1:
            raise AudioSignalException('Cannot support arrays with less than 2 dimensions!')

        if value.ndim == 2:
            value = np.expand_dims(value, axis=constants.STFT_CHAN_INDEX)

        if value.ndim > 3:
            raise AudioSignalException('Cannot support arrays with more than 3 dimensions!')

        if not np.iscomplexobj(value):
            warnings.warn('Initializing STFT with data that is non-complex. '
                          'This might lead to weird results!')

        self._stft_data = value

    @property
    def stft_params(self):
        """
        ``STFTParams``
            STFT parameters are kept in this property. STFT parameters are a ``namedtuple``
            called ``STFTParams`` with the following signature:

            .. code-block:: python

                STFTParams(
                    window_length=2048,
                    hop_length=512,
                    window_type='hann'
                )

            The defaults are 32ms windows, 8ms hop, and a hann window.

        """
        return self._stft_params

    @stft_params.setter
    def stft_params(self, value):
        if value and not isinstance(value, STFTParams):
            raise ValueError("stft_params must be of type STFTParams or None!")

        default_win_len = int(
            2 ** (np.ceil(np.log2(constants.DEFAULT_WIN_LEN_PARAM * self.sample_rate)))
        )
        default_hop_len = default_win_len // 4
        default_win_type = constants.WINDOW_DEFAULT

        default_stft_params = STFTParams(
            window_length=default_win_len,
            hop_length=default_hop_len,
            window_type=default_win_type
        )._asdict()

        value = value._asdict() if value else default_stft_params

        for key in default_stft_params:
            if value[key] is None:
                value[key] = default_stft_params[key]

        self._stft_params = STFTParams(**value)
        if self._stft_params.window_type == 'sqrt_hann':
            window_type = constants.WINDOW_HANN
        else:
            window_type = self._stft_params.window_type
        check_COLA(window_type, self._stft_params.window_length, self._stft_params.hop_length)

    @property
    def has_data(self):
        """
        ``bool``
            Returns ``False`` if :attr:`audio_data` and :attr:`stft_data` are empty. Else,
            returns ``True``.
        """
        has_audio_data = self.audio_data is not None and self.audio_data.size != 0
        has_stft_data = self.stft_data is not None and self.stft_data.size != 0
        return has_audio_data or has_stft_data

    @property
    def file_name(self):
        """
        ``str``
            The name of the file associated with this object. Includes extension, but not the full
            path.
        
        Notes:
            This will return ``None`` if this :class:`AudioSignal` object was not
            loaded from a file.
        
        See Also:
            :attr:`path_to_input_file` for the full path.
        """
        if self.path_to_input_file is not None:
            return os.path.basename(self.path_to_input_file)
        return None

    @property
    def sample_rate(self):
        """
        ``int``
            Sample rate associated with this object. If audio was read from a file, the sample
            rate will be set to the sample rate associated with the file. If this object was
            initialized from an array then the sample rate is set upon init. This property is
            read-only. To change the sample rate, use :func:`resample`.

        Notes:
            This property is read-only and cannot be set directly. To change

        See Also:
            * :func:`resample` to change the sample rate and resample data in :attr:`sample_rate`.

            * :func:`load_audio_from_array` to read audio from an array and set the sample rate.

            * :var:`nussl.constants.DEFAULT_SAMPLE_RATE` the default sample rate for *nussl*
                if not specified
        """
        return self._sample_rate

    @property
    def time_vector(self):
        """
        ``np.ndarray``
            A 1D :obj:`np.ndarray` with timestamps (in seconds) for each sample in
            :attr:`audio_data`.
        """
        if self.signal_duration is None:
            return None
        return np.linspace(0.0, self.signal_duration, num=self.signal_length)

    @property
    def freq_vector(self):
        """
        ``np.ndarray``
            A 1D numpy array with frequency values (in Hz) that correspond
            to each frequency bin (vertical axis) in :attr:`stft_data`. Assumes
            linearly spaced frequency bins.

        Raises:
            :class:`AudioSignalException`: If :attr:`stft_data` is ``None``. 
                Run :func:`stft` before accessing this.
        """
        if self.stft_data is None:
            raise AudioSignalException(
                'Cannot calculate freq_vector until self.stft() is run')
        return np.linspace(
            0.0, self.sample_rate // 2,
            num=self.stft_data.shape[constants.STFT_VERT_INDEX])

    @property
    def time_bins_vector(self):
        """
        ``np.ndarray``
            A 1D numpy array with time values (in seconds) that correspond
            to each time bin (horizontal/time axis) in :attr:`stft_data`.

        Raises:
            :class:`AudioSignalException`: If :attr:`stft_data` is ``None``. Run :func:`stft`
                before accessing this.
        """
        if self.stft_data is None:
            raise AudioSignalException(
                'Cannot calculate time_bins_vector until self.stft() is run')
        return np.linspace(0.0, self.signal_duration,
                           num=self.stft_data.shape[constants.STFT_LEN_INDEX])

    @property
    def stft_length(self):
        """
        ``int``
            The length of :attr:`stft_data` along the time axis. In units of hops.

        Raises:
            :class:`AudioSignalException`: If ``self.stft_dat``a is ``None``. Run :func:`stft`
                before accessing this.
        """
        if self.stft_data is None:
            raise AudioSignalException('Cannot calculate stft_length until self.stft() is run')
        return self.stft_data.shape[constants.STFT_LEN_INDEX]

    @property
    def active_region_is_default(self):
        """
        ``bool``
            ``True`` if active region is the full length of :attr:`audio_data`. ``False`` otherwise.

        See Also:

            * :func:`set_active_region` for a description of active regions in :class:`AudioSignal`

            * :func:`set_active_region_to_default`
        """
        return self._active_start == 0 and self._active_end == self._signal_length

    @property
    def _signal_length(self):
        """
        ``int``
            This is the length of the full signal, not just the active region.

        """
        if self._audio_data is None:
            return None
        return self._audio_data.shape[constants.LEN_INDEX]

    @property
    def power_spectrogram_data(self):
        """
        ``np.ndarray``
            Returns a real valued :obj:`np.ndarray` with power
            spectrogram data. The power spectrogram is defined as ``(STFT)^2``, where ``^2`` is
            element-wise squaring of entries of the STFT. Same shape as :attr:`stft_data`.
        
        Raises:
            :class:`AudioSignalException`: if :attr:`stft_data` is ``None``. Run :func:`stft`
                before accessing this.
            
        See Also:
            * :func:`stft` to calculate the STFT before accessing this attribute.
            * :attr:`stft_data` complex-valued Short-time Fourier Transform data.
            * :attr:`magnitude_spectrogram_data` to get magnitude spectrogram data.
            * :func:`get_power_spectrogram_channel` to get a specific channel
            
        """
        if self.stft_data is None:
            raise AudioSignalException('Cannot calculate power_spectrogram_data '
                                       'because self.stft_data is None')
        return np.abs(self.stft_data) ** 2

    @property
    def magnitude_spectrogram_data(self):
        """
        ``np.ndarray``
            Returns a real valued ``np.array`` with magnitude spectrogram data. The magnitude
            spectrogram is defined as ``abs(STFT)``, the element-wise absolute value of every item
            in the STFT. Same shape as :attr:`stft_data`.
        
        Raises:
            AudioSignalException: if :attr:`stft_data` is ``None``. Run :func:`stft` before
                accessing this.
            
        See Also:
            * :func:`stft` to calculate the STFT before accessing this attribute.
            * :attr:`stft_data` complex-valued Short-time Fourier Transform data.
            * :attr:`power_spectrogram_data`
            * :func:`get_magnitude_spectrogram_channel`
            
        """
        if self.stft_data is None:
            raise AudioSignalException('Cannot calculate magnitude_spectrogram_data '
                                       'because self.stft_data is None')
        return np.abs(self.stft_data)

    @property
    def log_magnitude_spectrogram_data(self):
        """
        (:obj:`np.ndarray`): Returns a real valued ``np.array`` with log magnitude spectrogram data.
        
        The log magnitude spectrogram is defined as 20 * log10(abs(stft)).
        Same shape as :attr:`stft_data`.
        
        Raises:
            AudioSignalException: if :attr:`stft_data` is ``None``. Run :func:`stft` before
                accessing this.
            
        See Also:
            * :func:`stft` to calculate the STFT before accessing this attribute.
            * :attr:`stft_data` complex-valued Short-time Fourier Transform data.
            * :attr:`power_spectrogram_data`
            * :func:`get_magnitude_spectrogram_channel`
            
        """
        if self.stft_data is None:
            raise AudioSignalException('Cannot calculate log_magnitude_spectrogram_data '
                                       'because self.stft_data is None')
        return 20 * np.log10(np.abs(self.stft_data) + 1e-8)
    
    @property
    def effects_chain(self):
        """
        (``list`` of ``nussl.core.FilterFunction``): Returns a copy of the AudioSignal's
        effect chain. Editing this property will not result in a change to the effects chain
        of the AudioSignal. 

        Please use the effects hooks (e.g. :func:`tremolo`, :func:`make_effect`) to make changes
        to the Audiosignal's effects chain.

        See Also:
            * :func:`apply_effects`
        """
        return self._effects_chain.copy()

    @property
    def effects_applied(self):
        """
        (``list`` of ``nussl.core.FilterFunction``): Returns a copy of the list of effects 
        applied to the AudioSignal. Editing this property will not result in a change to the 
        effects aplied to the AudioSignal. 

        Please use :func:`apply_effects` to apply effects to the AudioSignal.
         
        See Also:
            * :func:`apply_effects`
        """
        return self._effects_applied.copy()

    ##################################################
    #                     I/O
    ##################################################

    def load_audio_from_file(self, input_file_path, offset=0, duration=None, new_sample_rate=None):
        # type: (str, float, float, int) -> None
        """
        Loads an audio signal into memory from a file on disc. The audio is stored in
        :class:`AudioSignal` as a :obj:`np.ndarray` of `float` s. The sample rate is read from
        the file, and this :class:`AudioSignal` object's sample rate is set from it. If
        :param:`new_sample_rate` is not ``None`` nor the same as the sample rate of the file,
        the audio will be resampled to the sample rate provided in the :param:`new_sample_rate`
        parameter. After reading the audio data into memory, the active region is set to default.

        :param:`offset` and :param:`duration` allow the user to determine how much of the audio is
        read from the file. If those are non-default, then only the values provided will be stored
        in :attr:`audio_data` (unlike with the active region, which has the entire audio data stored
        in memory but only allows access to a subset of the audio).

        See Also:
            * :func:`load_audio_from_array` to read audio data from a :obj:`np.ndarray`.

        Args:
            input_file_path (str): Path to input file.
            offset (float,): The starting point of the section to be extracted (seconds).
                Defaults to 0 seconds (i.e., the very beginning of the file).
            duration (float): Length of signal to load in second.
                signal_length of 0 means read the whole file. Defaults to the full
                length of the signal.
            new_sample_rate (int): If this parameter is not ``None`` or the same sample rate as
                provided by the input file, then the audio data will be resampled to the new
                sample rate dictated by this parameter.

        """
        assert offset >= 0, 'Parameter `offset` must be >= 0!'
        if duration is not None:
            assert duration >= 0, 'Parameter `duration` must be >= 0!'

        try:
            # try reading headers with soundfile for speed
            audio_info = sf.info(input_file_path)
            file_length = audio_info.duration
        except:
            # if that doesn't work try audioread
            with audioread.audio_open(os.path.realpath(input_file_path)) as input_file:
                file_length = input_file.duration

        if offset > file_length:
            raise AudioSignalException('offset is longer than signal!')

        if duration is not None and offset + duration >= file_length:
            warnings.warn('offset + duration are longer than the signal.'
                          ' Reading until end of signal...',
                          UserWarning)

        audio_input, self._sample_rate = librosa.load(input_file_path,
                                                      sr=None,
                                                      offset=offset,
                                                      duration=duration,
                                                      mono=False)

        self.audio_data = audio_input
        self.original_signal_length = self.signal_length

        if new_sample_rate is not None and new_sample_rate != self._sample_rate:
            warnings.warn('Input sample rate is different than the sample rate'
                          ' read from the file! Resampling...',
                          UserWarning)
            self.resample(new_sample_rate)

        self.path_to_input_file = input_file_path
        self.set_active_region_to_default()

    def load_audio_from_array(self, signal, sample_rate=constants.DEFAULT_SAMPLE_RATE):
        """
        Loads an audio signal from a :obj:`np.ndarray`. :param:`sample_rate` is the sample
        of the signal.

        See Also:
            * :func:`load_audio_from_file` to read in an audio file from disc.

        Notes:
            Only accepts float arrays and int arrays of depth 16-bits.

        Parameters:
            signal (:obj:`np.ndarray`): Array containing the audio signal sampled at
                :param:`sample_rate`.
            sample_rate (int): The sample rate of signal.
                Default is :ref:`constants.DEFAULT_SAMPLE_RATE` (44.1kHz)

        """
        assert (type(signal) == np.ndarray)

        self.path_to_input_file = None

        # Change from fixed point to floating point
        if not np.issubdtype(signal.dtype, np.floating):
            signal = signal.astype('float') / (np.iinfo(np.dtype('int16')).max + 1.0)

        self.audio_data = signal
        self.original_signal_length = self.signal_length
        self._sample_rate = sample_rate if sample_rate is not None \
            else constants.DEFAULT_SAMPLE_RATE

        self.set_active_region_to_default()

    def write_audio_to_file(self, output_file_path, sample_rate=None):
        """
        Outputs the audio signal data in :attr:`audio_data` to a file at :param:`output_file_path`
        with sample rate of :param:`sample_rate`.

        Parameters:
            output_file_path (str): Filename where output file will be saved.
            sample_rate (int): The sample rate to write the file at. Default is
                :attr:`sample_rate`.
        """
        if self.audio_data is None:
            raise AudioSignalException("Cannot write audio file because there is no audio data.")

        if sample_rate is None:
            sample_rate = self.sample_rate

        audio_output = np.copy(self.audio_data)

        # TODO: better fix
        # convert to fixed point again
        if not np.issubdtype(audio_output.dtype, np.dtype(int).type):
            audio_output = np.multiply(
                audio_output,
                2 ** (constants.DEFAULT_BIT_DEPTH - 1)).astype('int16')
        wav.write(output_file_path, sample_rate, audio_output.T)

    ##################################################
    #                Active Region
    ##################################################

    def set_active_region(self, start, end):
        """
        Determines the bounds of what gets returned when you access :attr:`audio_data`.
        None of the data in :attr:`audio_data` is discarded when you set the active region, it
        merely becomes inaccessible until the active region is set back to default (i.e., the full
        length of the signal).

        This is useful for reusing a single :class:`AudioSignal` object to do multiple operations on
        only select parts of the audio data.

        Warnings:
            Many functions will raise exceptions while the active region is not default. Be aware
            that adding, subtracting, concatenating, truncating, and other utilities are not
            available when the active region is not default.

        See Also:
            * :func:`set_active_region_to_default`
            * :attr:`active_region_is_default`

        Examples:
            >>> import nussl
            >>> import numpy as np
            >>> n = nussl.constants.DEFAULT_SAMPLE_RATE  # 1 second of audio at 44.1kHz
            >>> np_sin = np.sin(np.linspace(0, 100 * 2 * np.pi, n))  # sine wave @ 100 Hz
            >>> sig = nussl.AudioSignal(audio_data_array=np_sin)
            >>> sig.signal_duration
            1.0
            >>> sig.set_active_region(0, n // 2)
            >>> sig.signal_duration
            0.5

        Args:
            start (int): Beginning of active region (in samples). Cannot be less than 0.
            end (int): End of active region (in samples). Cannot be larger than
                :attr:`signal_length`.

        """
        start, end = int(start), int(end)
        self._active_start = start if start >= 0 else 0
        self._active_end = end if end < self._signal_length else self._signal_length

    def set_active_region_to_default(self):
        """
        Resets the active region of this :class:`AudioSignal` object to its default value of the
        entire :attr:`audio_data` array.
        
        See Also:
            * :func:`set_active_region` for an explanation of active regions within the
            :class:`AudioSignal`.

        """
        self._active_start = 0
        self._active_end = self._signal_length

    ##################################################
    #               STFT Utilities
    ##################################################

    @staticmethod
    def get_window(window_type, window_length):
        """
        Wrapper around scipy.signal.get_window so one can also get the 
        popular sqrt-hann window.
        
        Args:
            window_type (str): Type of window to get (see constants.ALL_WINDOW).
            window_length (int): Length of the window
        
        Returns:
            np.ndarray: Window returned by scipy.signa.get_window
        """
        if window_type == constants.WINDOW_SQRT_HANN:
            window = np.sqrt(scipy.signal.get_window(
                'hann', window_length
            ))
        else:
            window = scipy.signal.get_window(
                window_type, window_length)

        return window

    def stft(self, window_length=None, hop_length=None, window_type=None, overwrite=True):
        """
        Computes the Short Time Fourier Transform (STFT) of :attr:`audio_data`.
        The results of the STFT calculation can be accessed from :attr:`stft_data`
        if :attr:`stft_data` is ``None`` prior to running this function or ``overwrite == True``

        Warning:
            If overwrite=True (default) this will overwrite any data in :attr:`stft_data`!

        Args:
            window_length (int): Amount of time (in samples) to do an FFT on
            hop_length (int): Amount of time (in samples) to skip ahead for the new FFT
            window_type (str): Type of scaling to apply to the window.
            overwrite (bool): Overwrite :attr:`stft_data` with current calculation

        Returns:
            (:obj:`np.ndarray`) Calculated, complex-valued STFT from :attr:`audio_data`, 3D numpy
            array with shape `(n_frequency_bins, n_hops, n_channels)`.

        """
        if self.audio_data is None or self.audio_data.size == 0:
            raise AudioSignalException(
                "No time domain signal (self.audio_data) to make STFT from!")

        window_length = (
            self.stft_params.window_length
            if window_length is None
            else int(window_length)
        )
        hop_length = (
            self.stft_params.hop_length
            if hop_length is None
            else int(hop_length)
        )
        window_type = (
            self.stft_params.window_type
            if window_type is None
            else window_type
        )

        stft_data = []

        window = self.get_window(window_type, window_length)

        for chan in self.get_channels():
            _, _, _stft = scipy.signal.stft(
                chan, fs=self.sample_rate, window=window,
                nperseg=window_length, noverlap=window_length - hop_length)
            stft_data.append(_stft)

        stft_data = np.array(stft_data).transpose((1, 2, 0))

        if overwrite:
            self.stft_data = stft_data

        return stft_data

    def istft(self, window_length=None, hop_length=None, window_type=None, overwrite=True,
              truncate_to_length=None):
        """ Computes and returns the inverse Short Time Fourier Transform (iSTFT).

        The results of the iSTFT calculation can be accessed from :attr:`audio_data`
        if :attr:`audio_data` is ``None`` prior to running this function or ``overwrite == True``

        Warning:
            If overwrite=True (default) this will overwrite any data in :attr:`audio_data`!

        Args:
            window_length (int): Amount of time (in samples) to do an FFT on
            hop_length (int): Amount of time (in samples) to skip ahead for the new FFT
            window_type (str): Type of scaling to apply to the window.
            overwrite (bool): Overwrite :attr:`stft_data` with current calculation
            truncate_to_length (int): truncate resultant signal to specified length. Default ``None``.

        Returns:
            (:obj:`np.ndarray`) Calculated, real-valued iSTFT from :attr:`stft_data`, 2D numpy array
            with shape `(n_channels, n_samples)`.

        """
        if self.stft_data is None or self.stft_data.size == 0:
            raise AudioSignalException('Cannot do inverse STFT without self.stft_data!')

        window_length = (
            self.stft_params.window_length
            if window_length is None
            else int(window_length)
        )
        hop_length = (
            self.stft_params.hop_length
            if hop_length is None
            else int(hop_length)
        )
        window_type = (
            self.stft_params.window_type
            if window_type is None
            else window_type
        )

        signals = []

        window = self.get_window(window_type, window_length)

        for stft in self.get_stft_channels():
            _, _signal = scipy.signal.istft(
                stft, fs=self.sample_rate, window=window,
                nperseg=window_length, noverlap=window_length - hop_length)

            signals.append(_signal)

        calculated_signal = np.array(signals)

        # Make sure it's shaped correctly
        calculated_signal = np.expand_dims(calculated_signal, -1) \
            if calculated_signal.ndim == 1 else calculated_signal

        # if truncate_to_length isn't provided
        if truncate_to_length is None:
            truncate_to_length = self.original_signal_length
            if self.signal_length is not None:
                truncate_to_length = self.signal_length

        if truncate_to_length is not None and truncate_to_length > 0:
            calculated_signal = calculated_signal[:, :truncate_to_length]

        if overwrite or self.audio_data is None:
            self.audio_data = calculated_signal

        return calculated_signal

    def apply_mask(self, mask, overwrite=False):
        """
        Applies the input mask to the time-frequency representation in this :class:`AudioSignal`
        object and returns a new :class:`AudioSignal` object with the mask applied. The mask
        is applied to the magnitude of audio signal. The phase of the original audio
        signal is then applied to construct the masked STFT.
        
        Args:
            mask (:obj:`MaskBase`-derived object): A ``MaskBase``-derived object 
                containing a mask.
            overwrite (bool): If ``True``, this will alter ``stft_data`` in self. 
                If ``False``, this function will create a new ``AudioSignal`` object 
                with the mask applied.

        Returns:
            A new :class:`AudioSignal`` object with the input mask applied to the STFT,
            iff ``overwrite`` is False.

        """
        if not isinstance(mask, masks.MaskBase):
            raise AudioSignalException(f'Expected MaskBase-derived object, given {type(mask)}')

        if self.stft_data is None:
            raise AudioSignalException('There is no STFT data to apply a mask to!')

        if mask.shape != self.stft_data.shape:
            if not mask.shape[:-1] == self.stft_data.shape[:-1]:
                raise AudioSignalException(
                    'Input mask and self.stft_data are not the same shape! mask:'
                    f' {mask.shape}, self.stft_data: {self.stft_data.shape}'
                )

        magnitude, phase = np.abs(self.stft_data), np.angle(self.stft_data)
        masked_abs = magnitude * mask.mask
        masked_stft = masked_abs * np.exp(1j * phase)

        if overwrite:
            self.stft_data = masked_stft
        else:
            return self.make_copy_with_stft_data(masked_stft, verbose=False)

    def ipd_ild_features(self, ch_one=0, ch_two=1):
        """
        Computes interphase difference (IPD) and interlevel difference (ILD) for a 
        stereo spectrogram. If more than two channels, this by default computes IPD/ILD
        between the first two channels. This can be specified by the arguments ch_one
        and ch_two. If only one channel, this raises an error.
        
        Args:
            ch_one (``int``): index of first channel to compute IPD/ILD.
            ch_two (``int``): index of second channel to compute IPD/ILD.

        Returns:
            ipd (``np.ndarray``): Interphase difference between selected channels
            ild (``np.ndarray``): Interlevel difference between selected channels

        """
        if self.stft_data is None:
            raise AudioSignalException("Cannot compute ipd/ild features without stft_data!")
        if self.is_mono:
            raise AudioSignalException("Cannot compute ipd/ild features on mono input!")

        stft_ch_one = self.get_stft_channel(ch_one)
        stft_ch_two = self.get_stft_channel(ch_two)

        ild = np.abs(stft_ch_one) / (np.abs(stft_ch_two) + 1e-4)
        ild = 20 * np.log10(ild + 1e-8)

        frequencies = self.freq_vector
        ipd = np.angle(stft_ch_two * np.conj(stft_ch_one))
        ipd /= (frequencies + 1.0)[:, None]
        ipd = ipd % np.pi

        return ipd, ild
    

    ##################################################
    #                  Utilities
    ##################################################

    def concat(self, other):
        """ Concatenate two :class:`AudioSignal` objects (by concatenating :attr:`audio_data`).

        Puts ``other.audio_data`` after :attr:`audio_data`.

        Raises:
            AudioSignalException: If ``self.sample_rate != other.sample_rate``,
                ``self.num_channels != other.num_channels``, or ``!self.active_region_is_default``
                is ``False``.

        Args:
            other (:class:`AudioSignal`): :class:`AudioSignal` to concatenate with the current one.
            
        """
        self._verify_audio(other)

        self.audio_data = np.concatenate((self.audio_data, other.audio_data),
                                         axis=constants.LEN_INDEX)

    def truncate_samples(self, n_samples):
        """ Truncates the signal leaving only the first ``n_samples`` samples.
        This can only be done if ``self.active_region_is_default`` is True. If
        ``n_samples > self.signal_length``, then `n_samples = self.signal_length` 
        (no truncation happens).

        Raises:
            AudioSignalException: If ``self.active_region_is_default`` is ``False``.

        Args:
            n_samples: (int) number of samples that will be left.

        """
        if not self.active_region_is_default:
            raise AudioSignalException('Cannot truncate while active region is not set as default!')

        n_samples = int(n_samples)
        if n_samples > self.signal_length:
            n_samples = self.signal_length

        self.audio_data = self.audio_data[:, 0: n_samples]

    def truncate_seconds(self, n_seconds):
        """ Truncates the signal leaving only the first n_seconds.
        This can only be done if self.active_region_is_default is True.

        Args:
            n_seconds: (float) number of seconds to truncate :attr:`audio_data`.

        """
        n_samples = int(n_seconds * self.sample_rate)
        self.truncate_samples(n_samples)

    def crop_signal(self, before, after):
        """
        Get rid of samples before and after the signal on all channels. Contracts the length
        of :attr:`audio_data` by before + after. Useful to get rid of zero padding after the fact.

        Args:
            before: (int) number of samples to remove at beginning of self.audio_data
            after: (int) number of samples to remove at end of self.audio_data

        """
        if not self.active_region_is_default:
            raise AudioSignalException('Cannot crop signal while active region '
                                       'is not set as default!')
        num_samples = self.signal_length
        self.audio_data = self.audio_data[:, before:num_samples - after]
        self.set_active_region_to_default()

    def zero_pad(self, before, after):
        """ Adds zeros before and after the signal to all channels.
        Extends the length of self.audio_data by before + after.

        Raises:
            Exception: If `self.active_region_is_default`` is ``False``.

        Args:
            before: (int) number of zeros to be put before the current contents of self.audio_data
            after: (int) number of zeros to be put after the current contents fo self.audio_data

        """
        if not self.active_region_is_default:
            raise AudioSignalException('Cannot zero-pad while active region is not set as default!')

        self.audio_data = np.pad(self.audio_data, ((0, 0), (before, after)), 'constant')

    def add(self, other):
        """Adds two audio signal objects.

        This does element-wise addition on the :attr:`audio_data` array.

        Raises:
            AudioSignalException: If ``self.sample_rate != other.sample_rate``,
                ``self.num_channels != other.num_channels``, or
                ``self.active_region_is_default`` is ``False``.

        Parameters:
            other (:class:`AudioSignal`): Other :class:`AudioSignal` to add.

        Returns:
            (:class:`AudioSignal`): New :class:`AudioSignal` object with the sum of
            ``self`` and ``other``.
        """
        if isinstance(other, int):
            # this is so that sum(list of audio_signals) works.
            # when sum is called on a list it's evaluated as 0 + elem1 + elem2 + ...
            # so the 0 case needs to be taken care of (by doing nothing)
            return self

        self._verify_audio_arithmetic(other)

        new_signal = copy.deepcopy(self)
        new_signal.audio_data = self.audio_data + other.audio_data

        return new_signal

    def subtract(self, other):
        """Subtracts two audio signal objects.

        This does element-wise subtraction on the :attr:`audio_data` array.

        Raises:
            AudioSignalException: If ``self.sample_rate != other.sample_rate``,
                ``self.num_channels != other.num_channels``, or
                ``self.active_region_is_default`` is ``False``.

        Parameters:
            other (:class:`AudioSignal`): Other :class:`AudioSignal` to subtract.

        Returns:
            (:class:`AudioSignal`): New :class:`AudioSignal` object with the difference
            between ``self`` and ``other``.
        """
        self._verify_audio_arithmetic(other)

        other_copy = copy.deepcopy(other)
        other_copy *= -1
        return self.add(other_copy)

    def make_copy_with_audio_data(self, audio_data, verbose=True):
        """ Makes a copy of this :class:`AudioSignal` object with :attr:`audio_data` initialized to
        the input :param:`audio_data` numpy array. The :attr:`stft_data` of the new
        :class:`AudioSignal`object is ``None``.

        Args:
            audio_data (:obj:`np.ndarray`): Audio data to be put into the new :class:`AudioSignal`
                object.
            verbose (bool): If ``True`` prints warnings. If ``False``, outputs nothing.

        Returns:
            (:class:`AudioSignal`): A copy of this :class:`AudioSignal` object with :attr:`audio_data`
            initialized to the input :param:`audio_data` numpy array.

        """
        if verbose:
            if not self.active_region_is_default:
                warnings.warn('Making a copy when active region is not default.')

            if audio_data.shape != self.audio_data.shape:
                warnings.warn('Shape of new audio_data does not match current audio_data.')

        new_signal = copy.deepcopy(self)
        new_signal.audio_data = audio_data
        new_signal.stft_data = None
        return new_signal

    def make_copy_with_stft_data(self, stft_data, verbose=True):
        """ Makes a copy of this :class:`AudioSignal` object with :attr:`stft_data` initialized to
        the input :param:`stft_data` numpy array. The :attr:`audio_data` of the new
        :class:`AudioSignal` object is ``None``.

        Args:
            stft_data (:obj:`np.ndarray`): STFT data to be put into the new :class:`AudioSignal`
                object.

        Returns:
            (:class:`AudioSignal`): A copy of this :class:`AudioSignal` object with :attr:`stft_data`
            initialized to the input :param:`stft_data` numpy array.

        """
        if verbose:
            if not self.active_region_is_default:
                warnings.warn('Making a copy when active region is not default.')

            if stft_data.shape != self.stft_data.shape:
                warnings.warn('Shape of new stft_data does not match current stft_data.')

        new_signal = copy.deepcopy(self)
        new_signal.stft_data = stft_data
        new_signal.original_signal_length = self.original_signal_length
        new_signal.audio_data = None
        return new_signal

    def loudness(self, filter_class='K-weighting', block_size=0.400):
        """
        Uses pyloudnorm to calculate loudness.

        Implementation of ITU-R BS.1770-4.
        Allows control over gating block size and frequency weighting filters for 
        additional control.

        Measure the integrated gated loudness of a signal.
        
        Uses the weighting filters and block size defined by the meter
        the integrated loudness is measured based upon the gating algorithm
        defined in the ITU-R BS.1770-4 specification. 
        Supports up to 5 channels and follows the channel ordering: 
        [Left, Right, Center, Left surround, Right surround]

        Args:
            filter_class (str):
              Class of weighting filter used.
              - 'K-weighting' (default)
              - 'Fenton/Lee 1'
              - 'Fenton/Lee 2'
              - 'Dash et al.'
            block_size (float):
              Gating block size in seconds. Defaults to 0.400.

        Returns:
            float: LUFS, Integrated gated loudness of the input 
              measured in dB LUFS.
        """

        # create BS.1770 meter
        meter = pyloudnorm.Meter(
            self.sample_rate, filter_class=filter_class, block_size=block_size)
        # measure loudness
        loudness = meter.integrated_loudness(self.audio_data.T) 
        return loudness

    def rms(self, win_len=None, hop_len=None):
        """ Calculates the root-mean-square of :attr:`audio_data`.
        
        Returns:
            (float): Root-mean-square of :attr:`audio_data`.

        """
        if win_len is not None:
            hop_len = win_len // 2 if hop_len is None else hop_len
            rms_func = lambda arr: librosa.feature.rms(arr, frame_length=win_len,
                                                       hop_length=hop_len)[0, :]
        else:
            rms_func = lambda arr: np.sqrt(np.mean(np.square(arr)))

        result = []
        for ch in self.get_channels():
            result.append(rms_func(ch))

        return np.squeeze(result)

    def peak_normalize(self):
        """
        Peak normalizes the audio signal.
        """
        self.apply_gain(1 / np.abs(self.audio_data).max())

    def apply_gain(self, value):
        """
        Apply a gain to :attr:`audio_data`

        Args:
            value (float): amount to multiply self.audio_data by

        Returns:
            (:class:`AudioSignal`): This :class:`AudioSignal` object with the gain applied.

        """
        if not isinstance(value, numbers.Real):
            raise AudioSignalException('Can only multiply/divide by a scalar!')

        self.audio_data = self.audio_data * value
        return self

    def resample(self, new_sample_rate, **kwargs):
        """
        Resample the data in :attr:`audio_data` to the new sample rate provided by
        :param:`new_sample_rate`. If the :param:`new_sample_rate` is the same as :attr:`sample_rate`
        then nothing happens.

        Args:
            new_sample_rate (int): The new sample rate of :attr:`audio_data`.
            kwargs: Keyword arguments to librosa.resample.

        """

        if new_sample_rate == self.sample_rate:
            warnings.warn('Cannot resample to the same sample rate.')
            return

        resampled_signal = []

        for channel in self.get_channels():
            resampled_channel = librosa.resample(
                channel, self.sample_rate, new_sample_rate, **kwargs)
            resampled_signal.append(resampled_channel)

        self.audio_data = np.array(resampled_signal)
        self.original_signal_length = self.signal_length
        self._sample_rate = new_sample_rate

    ##################################################
    #              Channel Utilities
    ##################################################

    def _verify_get_channel(self, n):
        if n >= self.num_channels:
            raise AudioSignalException(
                f'Cannot get channel {n} when this object only has {self.num_channels}'
                ' channels! (0-based)'
            )

        if n < 0:
            raise AudioSignalException(
                f'Cannot get channel {n}. This will cause unexpected results.'
            )

    def get_channel(self, n):
        """Gets audio data of n-th channel from :attr:`audio_data` as a 1D :obj:`np.ndarray`
        of shape ``(n_samples,)``.

        Parameters:
            n (int): index of channel to get. **0-based**

        See Also:
            * :func:`get_channels`: Generator for looping through channels of :attr:`audio_data`.
            * :func:`get_stft_channel`: Gets stft data from a specific channel.
            * :func:`get_stft_channels`: Generator for looping through channels from
            :attr:`stft_data`.

        Raises:
            :class:`AudioSignalException`: If not ``0 <= n < self.num_channels``.
            
        Returns:
            (:obj:`np.array`): The audio data in the n-th channel of the signal, 1D

        """
        self._verify_get_channel(n)

        return np.asfortranarray(utils._get_axis(self.audio_data, constants.CHAN_INDEX, n))

    def get_channels(self):
        """Generator that will loop through channels of :attr:`audio_data`.

        See Also:
            * :func:`get_channel`: Gets audio data from a specific channel.
            * :func:`get_stft_channel`: Gets stft data from a specific channel.
            * :func:`get_stft_channels`: Generator to loop through channels of :attr:`stft_data`.

        Yields:
            (:obj:`np.array`): The audio data in the next channel of this signal as a
            1D ``np.ndarray``.

        """
        for i in range(self.num_channels):
            yield self.get_channel(i)

    def get_stft_channel(self, n):
        """Returns STFT data of n-th channel from :attr:`stft_data` as a 2D ``np.ndarray``.

        Args:
            n: (int) index of stft channel to get. **0-based**

        See Also:
            * :func:`get_stft_channels`: Generator to loop through channels from :attr:`stft_data`.
            * :func:`get_channel`: Gets audio data from a specific channel.
            * :func:`get_channels`: Generator to loop through channels of :attr:`audio_data`.

        Raises:
            :class:`AudioSignalException`: If not ``0 <= n < self.num_channels``.

        Returns:
            (:obj:`np.array`): the STFT data in the n-th channel of the signal, 2D

        """
        if self.stft_data is None:
            raise AudioSignalException('Cannot get STFT data before STFT is calculated!')

        self._verify_get_channel(n)

        return utils._get_axis(self.stft_data, constants.STFT_CHAN_INDEX, n)

    def get_stft_channels(self):
        """Generator that will loop through channels of :attr:`stft_data`.

        See Also:
            * :func:`get_stft_channel`: Gets stft data from a specific channel.
            * :func:`get_channel`: Gets audio data from a specific channel.
            * :func:`get_channels`: Generator to loop through channels of :attr:`audio_data`.

        Yields:
            (:obj:`np.array`): The STFT data in the next channel of this signal as a
            2D ``np.ndarray``.

        """
        for i in range(self.num_channels):
            yield self.get_stft_channel(i)

    def make_audio_signal_from_channel(self, n):
        """
        Makes a new :class:`AudioSignal` object from with data from channel ``n``.
        
        Args:
            n (int): index of channel to make a new signal from. **0-based**

        Returns:
            (:class:`AudioSignal`) new :class:`AudioSignal` object with only data from
            channel ``n``.

        """
        new_signal = copy.copy(self)
        new_signal.audio_data = self.get_channel(n)
        return new_signal

    def get_power_spectrogram_channel(self, n):
        """ Returns the n-th channel from ``self.power_spectrogram_data``.

        Raises:
            Exception: If not ``0 <= n < self.num_channels``.

        Args:
            n: (int) index of power spectrogram channel to get **0-based**

        Returns:
            (:obj:`np.array`): the power spectrogram data in the n-th channel of the signal, 1D
        """
        self._verify_get_channel(n)

        # np.array helps with duck typing
        return utils._get_axis(np.array(self.power_spectrogram_data),
                               constants.STFT_CHAN_INDEX, n)

    def get_magnitude_spectrogram_channel(self, n):
        """ Returns the n-th channel from ``self.magnitude_spectrogram_data``.

        Raises:
           Exception: If not ``0 <= n < self.num_channels``.

        Args:
            n: (int) index of magnitude spectrogram channel to get **0-based**

        Returns:
            (:obj:`np.array`): the magnitude spectrogram data in the n-th channel of the signal, 1D
        """
        self._verify_get_channel(n)

        # np.array helps with duck typing
        return utils._get_axis(np.array(self.magnitude_spectrogram_data),
                               constants.STFT_CHAN_INDEX, n)

    def to_mono(self, overwrite=True, keep_dims=False):
        """ Converts :attr:`audio_data` to mono by averaging every sample.

        Args:
            overwrite (bool): If ``True`` this function will overwrite :attr:`audio_data`.
            keep_dims (bool): If ``False`` this function will return a 1D array,
                else will return array with shape `(1, n_samples)`.

        Warning:
            If ``overwrite=True`` (default) this will overwrite any data in :attr:`audio_data`!

        Returns:
            (:obj:`AudioSignal`): Mono-ed version of AudioSignal, either in place or not.

        """
        mono = np.mean(self.audio_data, axis=constants.CHAN_INDEX, keepdims=keep_dims)

        if overwrite:
            self.audio_data = mono
            return self
        else:
            mono_signal = self.make_copy_with_audio_data(mono)
            return mono_signal

    ##################################################
    #                 Utility hooks                  #
    ##################################################

    def play(self):
        """
        Plays this audio signal, using `nussl.play_utils.play`.

        Plays an audio signal if ffplay from the ffmpeg suite of tools is installed.
        Otherwise, will fail. The audio signal is written to a temporary file
        and then played with ffplay.
        """
        # lazy load
        from . import play_utils
        play_utils.play(self)

    def embed_audio(self, ext='.mp3', display=True):
        """
        Embeds the audio signal into a notebook, using `nussl.play_utils.embed_audio`.
        
        Write a numpy array to a temporary mp3 file using ffmpy, then embeds the mp3 
        into the notebook.

        Args:
            ext (str): What extension to use when embedding. '.mp3' is more lightweight 
            leading to smaller notebook sizes.
            display (bool): Whether or not to display the object immediately, or to return
            the html object for display later by the end user. Defaults to True.

        Example:
            >>> import nussl
            >>> audio_file = nussl.efz_utils.download_audio_file('schoolboy_fascination_excerpt.wav')
            >>> audio_signal = nussl.AudioSignal(audio_file)
            >>> audio_signal.embed_audio()

        This will show a little audio player where you can play the audio inline in 
        the notebook.        
        """
        # lazy load
        from . import play_utils
        return play_utils.embed_audio(self, ext=ext, display=display)

    ##################################
    #          Effect Hooks          #
    ##################################

    def apply_effects(self, reset=True, overwrite=False, user_order=True):
        """
        This method applies a prespecified set of audio effects (e.g., chorus, filtering,
        reverb, etc...) to this audio signal. Before any effect can be applied, the effects
        are first added to the "effects chain", which refers to a queue of effects that will be
        all applied to an AudioSignal object when this function is called. Effects are added
        to the effects chain through AudioSignal effect hooks which are `AudioSignal` methods for
        setting up an effect with the desired parameters. If the effect chain is empty this method
        does nothing. By default, when this method is called the effects chain empty. See the
        documentation below for a list of supported effects and their respective details.

        Notes:
            The effects will be added in the order that they are added to the effects chain, unless
            `user_order=False`, in case the order is not guaranteed to be preserved. Setting
            `user_order=False` will apply all SoX effects first, then FFMpeg effects, which can
            sped up processing time by ~30% in our experiments.

        Args:
            reset (bool): If True, clears out all effects in effect chains following applying the
                effects. Default=True
            overwrite (bool): If True, overwrites existing audio_data in AudioSignal. Default=False
                Also clears out `stft_data`.
            user_order (bool): If True, applied effects in the user provided order. If False,
                applies all SoX effects before all FFmpeg effects, which can be faster.
        Returns:
            self or new_signal (AudioSignal): If overwrite=True, then returns initially AudioSignal
            with edited audio_data. Otherwise, returns a new AudioSignal new_signal.

        Example:
            Here are some examples of demonstrating to apply effects to your audio signal. Let's
            start with an obvious effect, such as time stretching. We can add
            this effect to the effects chain by using the built-in effects hook, `time_stretch()`:

            >>> signal.signal_duration
            10.0
            >>> signal.time_stretch(0.5)
            >>> signal.signal_duration
            10.0

            You can find this effect in the AudioSignal's effects chain. 

            >>> effect = signal.effects_chain[0]
            >>> str(effect)
            "time_stretch (params: {factor=0.5})"

            However, the signal's duration hasn't changed! You will need to call `apply_effects()`
            to apply the changes in the signal's effects chains. Applied effects can be found in 
            `effects_applied`.

            >>> new_signal = signal.apply_effects()
            >>> new_signal.signal_duration
            5.0
            >>> str(new_signal.effects_applied[0])
            "time_stretch (params: {factor=0.5})"

            >>> # This doesn't change the original signal
            >>> signal.signal_duration
            10.0

            You can iterate through effects_chain to use the properties of FilterFunction
            objects as arguments to `make_effect`:

            >>> for effect in signal1.effects_applied:
            >>>     filter_ = effect.filter
            >>>     params = effect.params
            >>>     signal2.make_effect(filter_, **params)

            Using `apply_effects()` will clear out the current effects chain. This behavior can be
            avoided by setting `reset` to False. 

            >>> another_signal = signal.apply_effects()
            >>> another_signal.signal_duration
            10.0

            To clear out the current effects chain without applying effect, use
            `reset_effects_chain()`. It will not revert effects already applied (i.e., your audio
            will still have the effects you applied).

            If `apply_effects()` is called with empty effects chain, then it returns itself.

            >>> another_signal == signal
            True

            You can also chain effects together. This will add a tremolo effect followed by a high
            pass filter effect to the AudioSignal's effects chain (Note: order matters!):

            >>> audio_signal.tremolo(5, .6).high_pass(12000)

            Using overwrite here, we change the audio data of the variable `audio_signal`, rather
            than create a new signal:

            >>> audio_signal.apply_effects(overwrite=True)
            >>> audio_signal.effects_applied
            ["tremolo", "highpass"]

            If `user_order` is false, FFmpeg effects will be applied AFTER SoX effects, regardless
            of the order the hooks are applied. The effects `time_stretch` and `pitch_shift` are SoX
            effects. All others are FFmpeg effects. This may be done for speed, as applying all
            FFmpeg effects without interuption will be faster than being interrupted with a SoX
            effect.

            For example, the two statements will result in the same effected signal:

            >>> signal_1 = audio_signal.pitch_shift(4).tremolo(5, .6).apply_effects(user_order=False)
            >>> signal_2 = audio_signal.tremolo(5, .6).pitch_shift(4).apply_effects(user_order=False)
            >>> signal_1.effects_applied == signal_2.effects_applied
            True

            Refer to the specific documentation for each effect to determine whether it is a SoX
            effect or an FFmpeg effect.

        See Also:
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
            * :func:`time_stretch`: Changes the length without effecting the pitch.
            * :func:`pitch_shift`: Changes the pitch without effecting the length of the signal.
            * :func:`low_pass`: Applies a low pass filter to the signal.
            * :func:`high_pass`: Applies a high pass filter to the signal.
            * :func:`tremelo`: Applies a tremolo (volume wobbling) effect to the signal.
            * :func:`vibrato`: Applies a vibrato (pitch wobbling) effect to the signal.
            * :func:`chorus`: Applies a chorus effect to the signal.
            * :func:`phaser`: Applies a phaser effect to the signal.
            * :func:`flanger`: Applies a flanger effect to the signal.
            * :func:`emphasis`: Boosts certain frequency ranges of the signal.
            * :func:`compressor`: Compresses the dynamic range of the signal.
            * :func:`equalizer`: Applies an equalizer to the signal.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
        """
        if user_order:
            new_signal = self._apply_user_ordered_effects()
        else:
            new_signal = self._apply_sox_ffmpeg_ordered_effects()
        new_signal.reset_effects_chain()
        if reset:
            self.reset_effects_chain()
        if overwrite:
            self.audio_data = new_signal.audio_data
            self._effects_applied += new_signal.effects_applied
            self.stft_data = None
            return self

        return new_signal

    def _apply_user_ordered_effects(self):
        new_signal = self
        i = j = 0

        while i < len(self._effects_chain):
            j += 1

            if j == len(self._effects_chain) or \
                type(self._effects_chain[i]) != type(self._effects_chain[j]):  # new fx type

                next_chain = self._effects_chain[i:j]
                if isinstance(next_chain[0], effects.SoXFilter):
                    new_signal = effects.apply_effects_sox(new_signal, next_chain)
                elif isinstance(next_chain[0], effects.FFmpegFilter):
                    new_signal = effects.apply_effects_ffmpeg(new_signal, next_chain)

                i = j

        return new_signal

    def _apply_sox_ffmpeg_ordered_effects(self):
        new_signal = self
        sox_effects_chain = []
        ffmpeg_effects_chain = []
        for f in self._effects_chain:
            if isinstance(f, effects.FFmpegFilter):
                ffmpeg_effects_chain.append(f)
            elif isinstance(f, effects.SoXFilter):
                sox_effects_chain.append(f)

        if sox_effects_chain:
            new_signal = effects.apply_effects_sox(new_signal, sox_effects_chain)
        if ffmpeg_effects_chain:
            new_signal = effects.apply_effects_ffmpeg(new_signal, ffmpeg_effects_chain)

        return new_signal

    def make_effect(self, effect, **kwargs):
        """
        Syntactic sugar for adding an arbitrary effect hook to the effects chain by name.

        Example:
            >>> signal.time_stretch(1.5)

            Is the same as
            >>> signal.make_effect("time_stretch", factor=1.5)

            The attributes of a FilterFunction in the lists effects_applied or effects_chain may 
            used with `make_effect`. 

            >>> for effect in signal1.effects_applied:
            >>>     filter_ = effect.filter
            >>>     params = effect.params
            >>>     signal2.make_effect(filter_, **params)

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args:
            effect (str): Function name of desired effect hook of the AudioSignal
            **kwargs: Additional parameters for given effect. 
        Return:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        try:
            effect_hook = getattr(self, effect, None)
            effect_hook(**kwargs)
        except Exception as e:
            raise AudioSignalException(f"Error calling {effect} with parameters {kwargs}: `{e}`")
        
        return self

    def reset_effects_chain(self):
        """
        Clears effects chain of AudioSignal.

        This will not revert effects that have already been applied to the audio!

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
        """
        self._effects_chain = []
        return self

    def time_stretch(self, factor, **kwargs):
        """
        Adds a time stretch filter to the AudioSignal's effects chain.
        A factor greater than one will shorten the signal, a factor less then one
        will lengthen the signal, and a factor of 1 will not change the signal.

        This is a SoX effect. Please see 
        https://pysox.readthedocs.io/en/latest/_modules/sox/transform.html#Transformer.tempo
        for details. 

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args: 
            factor (float): Scaling factor for tempo change. Must be positive.
            kwargs: Arugments passed to `sox.transform.tempo`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.time_stretch(factor, **kwargs))
        return self
    
    def pitch_shift(self, n_semitones, **kwargs):
        """
        Add pitch shift effect to AudioSignal's effect chain. 
        A positive shift will change the pitch of the signal by `n_semitones`
        semitones. If positive, pitch will get higher, if negative pitch will
        get lower.

        This is a SoX effect. Please see:
        https://pysox.readthedocs.io/en/latest/_modules/sox/transform.html#Transformer.pitch
        For details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args: 
            n_semitones (float): The number of semitones to shift the audio.
                Positive values increases the frequency of the signal
            kwargs: Arugments passed to `sox.transform.pitch`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.pitch_shift(n_semitones, **kwargs))
        return self
    
    def low_pass(self, freq, poles=2, width_type="h", width=0.707, **kwargs):
        """
        Add low pass effect to AudioSignal's effect chain

        This is a FFmpeg effect. Please see:
        https://ffmpeg.org/ffmpeg-all.html#lowpass
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!
        
        Args: 
            freq (float): Threshold for low pass. Should be positive
            poles (int): Number of poles. should be either 1 or 2
            width_type (str): Unit of width for filter. Must be either:
                'h': Hz
                'q': Q-factor
                'o': octave
                's': slope
                'k': kHz
            width (float): Band width in width_type units
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.low_pass(freq, poles=poles,
                                                    width_type=width_type,
                                                    width=width, **kwargs))
        return self
    
    def high_pass(self, freq, poles=2, width_type="h", width=0.707, **kwargs):
        """
        Add high pass effect to AudioSignal's effect chain

        This is a FFmpeg effect. Please see:
        https://ffmpeg.org/ffmpeg-all.html#highpass
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!
        
        Args: 
            freq (float): Threshold for high pass. Should be positive scalar
            poles (int): Number of poles. should be either 1 or 2
            width_type (str): Unit of width for filter. Must be either:
                'h': Hz
                'q': Q-factor
                'o': octave
                's': slope
                'k': kHz
            width (float): Band width in width_type units
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.high_pass(freq, poles=poles,
                                                     width_type=width_type,
                                                     width=width, **kwargs))
        return self

    def tremolo(self, mod_freq, mod_depth, **kwargs):
        """
        Add tremolo effect to AudioSignal's effect chain
        
        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#tremolo
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!
        
        Args: 
            mod_freq (float): Modulation frequency. Must be between .1 and 20000.
            mod_depth (float): Modulation depth. Must be between 0 and 1.
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.tremolo(mod_freq, mod_depth, **kwargs))
        return self
    
    def vibrato(self, mod_freq, mod_depth, **kwargs):
        """
        Add vibrato effect to AudioSignal's effect chain.

        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#vibrato
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!
        
        Args: 
            mod_freq (float): Modulation frequency. Must be between .1 and 20000.
            mod_depth (float): Modulation depth. Must be between 0 and 1.
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.vibrato(mod_freq, mod_depth, **kwargs))
        return self
    
    def chorus(self, delays, decays, speeds,
               depths, in_gain=0.4, out_gain=0.4, **kwargs):
        """
        Add chorus effect to AudioSignal's effect chain.

        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#chorus
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!
        
        Args:
            delays (list of float): delays in ms. Typical Delay is 40ms-6ms
            decays (list of float): decays. Must be between 0 and 1
            speeds (list of float): speeds. Must be between 0 and 1
            depths (list of float): depths. Must be between 0 and 1
            in_gain (float): Proportion of input gain. Must be between 0 and 1
            out_gain (float): Proportion of output gain. Must be between 0 and 1
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.chorus(delays, decays,
                                                  speeds, depths,
                                                  in_gain=in_gain,
                                                  out_gain=out_gain,
                                                  **kwargs))
        return self
        
    def phaser(self, in_gain=0.4, out_gain=0.74, delay=3, decay=0.4,
               speed=0.5, type_="triangular", **kwargs):
        """
        Add phaser effect to AudioSignal's effect chain

        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#aphaser
        for details. 

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args:
            in_gain (float): Proportion of input gain. Must be between 0 and 1
            out_gain (float): Proportion of output gain. Must be between 0 and 1.
            delay (float): Delay of chorus filter in ms. (Time between original signal and delayed)
            decay (float): Decay of copied signal. Must be between 0 and 1.
            speed (float): Modulation speed of the delayed filter. 
            type_ (str): modulation type. Either Triangular or Sinusoidal
                "triangular" or "t" for Triangular
                "sinusoidal" of "s" for sinusoidal
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        fx = effects.phaser(in_gain=in_gain, out_gain=out_gain, delay=delay,
                            decay=decay, speed=speed, type_=type_, **kwargs)
        self._effects_chain.append(fx)
        return self
    
    def flanger(self, delay=0, depth=2, regen=0, width=71, speed=0.5,
                phase=25, shape="sinusoidal", interp="linear", **kwargs):
        """
        Add flanger effect to AudioSignal's effect chain
        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#flanger
        for details. 

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args:
            delay (float): Base delay in ms between original signal and copy.
                Must be between 0 and 30.
            depth (float): Sweep delay in ms. Must be between 0 and 10.
            regen (float): Percentage regeneration, or delayed signal feedback.
                Must be between -95 and 95.
            width (float): Percentage of delayed signal. Must be between 0 and 100.
            speed (float): Sweeps per second. Must be in .1 to 10
            shape (str): Swept wave shape, Must be "triangular" or "sinusoidal".
            phase (float): swept wave percentage-shift for multi channel. Must be between 0 and 100.
            interp (str): Delay Line interpolation. Must be "linear" or "quadratic".
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        fx = effects.flanger(delay=delay, depth=depth, regen=regen, width=width,
                             speed=speed, phase=phase, shape=shape, interp=interp,
                             **kwargs)
        self._effects_chain.append(fx)
        return self
    
    def emphasis(self, level_in, level_out, type_="col", mode='production', **kwargs):
        """
        Add emphasis effect to AudioSignal's effect chain. An emphasis filter boosts 
        frequency ranges the most susceptible to noise in a medium. When restoring
        sounds from such a medium, a de-emphasis filter is used to de-boost boosted 
        frequencies. 

        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#aemphasis
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args:
            level_in (float): Input gain
            level_out (float): Output gain
            type_ (str): physical medium type to convert/deconvert from.
                Must be one of the following: 
                - "col": Columbia 
                - "emi": EMI
                - "bsi": BSI (78RPM)
                - "riaa": RIAA
                - "cd": CD (Compact Disk)
                - "50fm": 50µs FM
                - "75fm": 75µs FM 
                - "50kf": 50µs FM-KF 
                - "75kf": 75µs FM-KF 
            mode (str): Filter mode. Must be one of the following:
                - "reproduction": Apply de-emphasis filter
                - "production": Apply emphasis filter
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.emphasis(level_in, level_out,
                                                    type_=type_, mode=mode,
                                                    **kwargs))
        return self
    
    def compressor(self, level_in, mode="downward", reduction_ratio=2,
                   attack=20, release=250, makeup=1, knee=2.8284, link="average",
                   detection="rms", mix=1, threshold=0.125, **kwargs):
        """
        Add compressor effect to AudioSignal's effect chain
    
        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#acompressor
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args:
            level_in (float): Input Gain
            mode (str): Mode of compressor operation. Can either be "upward" or "downward". 
            threshold (float): Volume threshold. If a signal's volume is above the threshold,
                gain reduction would apply.
            reduction_ratio (float): Ratio in which the signal is reduced.
            attack (float): Time in ms between when the signal rises above threshold and when 
                reduction is applied
            release (float): Time in ms between when the signal fall below threshold and 
                when reduction is decreased.
            makeup (float): Factor of amplification post-processing
            knee (float): Softens the transition between reduction and lack of thereof. 
                Higher values translate to a softer transition. 
            link (str): Choose average between all channels or mean. String of either
                "average" or "mean.
            detection (str): Whether to process exact signal of the RMS of nearby signals. 
                Either "peak" for exact or "rms".
            mix (float): Proportion of compressed signal in output.
            kwargs: Arguments passed to `ffmpeg.filter`
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        fx = effects.compressor(level_in, mode=mode, reduction_ratio=reduction_ratio,
                                attack=attack, release=release, makeup=makeup, knee=knee, link=link,
                                detection=detection, mix=mix, threshold=threshold, **kwargs)
        self._effects_chain.append(fx)
        return self

    def equalizer(self, bands, **kwargs):
        """
        Add eqaulizer effect to AudioSignal's effect chain

        This is a FFmpeg effect. Please see
        https://ffmpeg.org/ffmpeg-all.html#anequalizer  
        for details.

        Notes:
            This effect won't be applied until you call `apply_effect()`!

        Args:
            bands: A list of dictionaries, for each band. The required values for each dictionary:
                'chn': List of channel numbers to apply filter. Must be list of ints.
                'f': central freqency of band
                'w': Width of the band in Hz
                'g': Band gain in dB
                't': Set filter type for band, optional, can be:
                    0, for Butterworth
                    1, for Chebyshev type 1
                    2, for Chebyshev type 2
        Returns:
            self: Initial AudioSignal with updated effect chains

        See Also:
            * :func:`apply_effects`: Applies effects once they are in the effects chain.
            * :func:`make_effect`: Syntactic sugar for adding an effect to the chain by name.
            * :func:`reset_effects_chain`: Empties the effects chain without applying any effects.
        """
        self._effects_chain.append(effects.equalizer(bands, **kwargs))
        return self

    ##################################################
    #              Operator overloading              #
    ##################################################

    def __add__(self, other):
        return self.add(other)

    def __radd__(self, other):
        return self.add(other)

    def __sub__(self, other):
        return self.subtract(other)

    def _verify_audio(self, other):
        if self.num_channels != other.num_channels:
            raise AudioSignalException('Cannot do operation with two signals that have '
                                       'a different number of channels!')

        if self.sample_rate != other.sample_rate:
            raise AudioSignalException('Cannot do operation with two signals that have '
                                       'different sample rates!')

    def _verify_audio_arithmetic(self, other):
        self._verify_audio(other)

        if self.signal_length != other.signal_length:
            raise AudioSignalException('Cannot do arithmetic with signals of different length!')

    def __iadd__(self, other):
        return self + other

    def __isub__(self, other):
        return self - other

    def __mul__(self, value):
        if not isinstance(value, numbers.Real):
            raise AudioSignalException('Can only multiply/divide by a scalar!')

        return self.make_copy_with_audio_data(np.multiply(self.audio_data, value), verbose=False)

    def __div__(self, value):
        if not isinstance(value, numbers.Real):
            raise AudioSignalException('Can only multiply/divide by a scalar!')

        return self.make_copy_with_audio_data(np.divide(self.audio_data, float(value)),
                                              verbose=False)

    def __truediv__(self, value):
        return self.__div__(value)

    def __itruediv__(self, value):
        return self.__idiv__(value)

    def __imul__(self, value):
        return self.apply_gain(value)

    def __idiv__(self, value):
        return self.apply_gain(1 / float(value))

    def __len__(self):
        return self.signal_length

    def __eq__(self, other):
        for k, v in list(self.__dict__.items()):
            if isinstance(v, np.ndarray):
                if not np.array_equal(v, other.__dict__[k]):
                    return False
            elif v != other.__dict__[k]:
                return False
        return True

    def __ne__(self, other):
        return not self == other


class AudioSignalException(Exception):
    """
    Exception class for :class:`AudioSignal`.
    """
    pass



"""
Evaluation
==========

Evaluation base
---------------

.. autoclass:: nussl.evaluation.EvaluationBase
    :members:
    :autosummary:

BSS Evaluation base
-------------------

.. autoclass:: nussl.evaluation.BSSEvaluationBase
    :members:
    :autosummary:

Scale invariant BSSEval
-----------------------

.. autoclass:: nussl.evaluation.BSSEvalScale
    :members:
    :autosummary:

.. autofunction:: nussl.evaluation.scale_bss_eval

BSSEvalV4 (museval)
-------------------

.. autoclass:: nussl.evaluation.BSSEvalV4
    :members:
    :autosummary:

Precision and recall on masks
-----------------------------

.. autoclass:: nussl.evaluation.PrecisionRecallFScore
    :members:
    :autosummary:

Aggregators
-----------

.. autofunction:: nussl.evaluation.aggregate_score_files


"""

from .report_card import aggregate_score_files, report_card, associate_metrics
from .evaluation_base import EvaluationBase
from .bss_eval import BSSEvaluationBase, BSSEvalV4, BSSEvalScale, scale_bss_eval
from .precision_recall_fscore import PrecisionRecallFScore



from ..base import MaskSeparationBase, SeparationException
from ...datasets import transforms


class IdealBinaryMask(MaskSeparationBase):
    """
    Implements an ideal binary mask (IBM) that is computed by using the known
    ground truth performance. This is one of the upper baselines.
    
    Args:
        input_audio_signal (AudioSignal): Signal to separate.
        sources (list): List of audio signal objects that correspond to the sources.
        mask_type (str, optional): Mask type. Defaults to 'binary'.
    """

    def __init__(self, input_audio_signal, sources, mask_type='binary',
                 mask_threshold=.5):
        if isinstance(sources, list):
            sources = {i: sources[i] for i in range(len(sources))}
        elif not isinstance(sources, dict):
            raise SeparationException("sources must be a list or a dict!")

        self.sources = sources
        super().__init__(
            input_audio_signal=input_audio_signal, mask_type=mask_type,
            mask_threshold=mask_threshold)

    def run(self):
        # Set up dictionary to pass to the transform.    
        data = {
            'mix': self.audio_signal,
            'sources': self.sources
        }

        msa = transforms.MagnitudeSpectrumApproximation()
        ibm = msa(data)['ideal_binary_mask']

        masks = []

        for i in range(ibm.shape[-1]):
            mask = self.mask_type(ibm[..., i])
            masks.append(mask)

        self.result_masks = masks
        return self.result_masks



from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('extra_requirements.txt') as f:
    extra_requirements = f.read().splitlines()

setup(
    name='nussl',
    version='1.1.3rc4',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Topic :: Artistic Software',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Editors',
        'Topic :: Software Development :: Libraries',
    ],
    description='A flexible sound source separation library.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='E. Manilow, P. Seetharaman, F. Pishdadian, N. Shelly, A. Bugler, B. Pardo',
    author_email='ethanmanilow@u.northwestern.edu',
    maintainer='E. Manilow, P. Seetharaman, F. Pishdadian, N. Shelly, A. Bugler, B. Pardo',
    maintainer_email='ethanmanilow@u.northwestern.edu',
    url='https://github.com/interactiveaudiolab/nussl',
    license='MIT',
    packages=find_packages(),
    package_data={'': ['core/templates/multitrack.html']},
    keywords=['audio', 'source', 'separation', 'music', 'sound', 'source separation'],
    install_requires=requirements,
    extras_require={
        'melodia': [
            'vamp'
        ],
        'extras': extra_requirements,
    }
)




"""
Cluster by inter-phase and inter-level difference
-------------------------------------------------

.. autoclass:: nussl.separation.spatial.SpatialClustering
    :autosummary:

PROJET: Separate via spatial projections
-------------------------------------------------

.. autoclass:: nussl.separation.spatial.Projet
    :autosummary:

DUET
----

.. autoclass:: nussl.separation.spatial.Duet
    :autosummary:

"""

from .spatial_clustering import SpatialClustering
from .projet import Projet
from .duet import Duet
